"""
Translate vehicle motion into global coordinates
requires output from vehicle Model
translates vehicle motion to global DataFrame
X - positive to the right
Y - positive up
vin - frome vehicle pre premotion or from impact model
delta_rad (steer tire turn angle)
beta_rad (velocity vector)
Dx, Dy, vehicle displacement in global frame

"""

import pandas as pd
import math

v_info = pd.read_excel('Input.xlsx', sheet_name = 'vehicles', header = 0)

#%% X,Y dataframes for drawing vehicle
head = pd.read_excel('Input.xlsx',sheet_name = 'draw_columns')
columns = head.columns.values.tolist() # Column Headers

def global_frame_df(vin, vehi):
    # convert dataframe of vehicle info to a dictionary for the designated vehicle
    if vehi == 1:
        v_dict = v_info[['label', 'v1']].copy().set_index('label').to_dict('dict')
        v_dict = v_dict['v1']
    if vehi == 2:
        v_dict = v_info[['label', 'v2']].copy().set_index('label').to_dict('dict')
        v_dict = v_dict['v2']

    draw_vx = pd.DataFrame(columns = columns)                                      # time variant vehicle geometry for plotting in vehicle frame
    draw_vy = pd.DataFrame(columns = columns)                                      # time variant vehicle geometry for plotting in vehicle frame

    # loop for creating seperate dataframes for X and Y components of vehicle drawing in vehicle frame
    for i in (range(len(vin))):

        draw_vx = draw_vx.append({'b_lfc': v_dict['lcgf'] + v_dict['f_hang'],     # body outline
                                  'b_rfc': v_dict['lcgf'] + v_dict['f_hang'],
                                  'b_rrc': -1* (v_dict['lcgr'] + v_dict['r_hang']),
                                  'b_lrc': -1* (v_dict['lcgr'] + v_dict['r_hang']),
                                  'lfw': v_dict['lcgf'],                         # left front wheel
                                  'lfw_a': v_dict['lcgf'] + v_dict['tire_d']/2*math.cos(vin.loc[i,'delta_rad']) - -1*v_dict['tire_w']/2*math.sin(vin.loc[i,'delta_rad']),
                                  'lfw_b': v_dict['lcgf'] + v_dict['tire_d']/2*math.cos(vin.loc[i,'delta_rad']) - v_dict['tire_w']/2*math.sin(vin.loc[i,'delta_rad']),
                                  'lfw_c': v_dict['lcgf'] + -1*v_dict['tire_d']/2*math.cos(vin.loc[i,'delta_rad']) - v_dict['tire_w']/2*math.sin(vin.loc[i,'delta_rad']),
                                  'lfw_d': v_dict['lcgf'] + -1*v_dict['tire_d']/2*math.cos(vin.loc[i,'delta_rad']) - -1*v_dict['tire_w']/2*math.sin(vin.loc[i,'delta_rad']),
                                  'rfw': v_dict['lcgf'],                         # Right front wheel
                                  'rfw_a': v_dict['lcgf'] + v_dict['tire_d']/2*math.cos(vin.loc[i,'delta_rad']) - -1*v_dict['tire_w']/2*math.sin(vin.loc[i,'delta_rad']),
                                  'rfw_b': v_dict['lcgf'] + v_dict['tire_d']/2*math.cos(vin.loc[i,'delta_rad']) - v_dict['tire_w']/2*math.sin(vin.loc[i,'delta_rad']),
                                  'rfw_c': v_dict['lcgf'] + -1*v_dict['tire_d']/2*math.cos(vin.loc[i,'delta_rad']) - v_dict['tire_w']/2*math.sin(vin.loc[i,'delta_rad']),
                                  'rfw_d': v_dict['lcgf'] + -1*v_dict['tire_d']/2*math.cos(vin.loc[i,'delta_rad']) - -1*v_dict['tire_w']/2*math.sin(vin.loc[i,'delta_rad']),
                                  'rrw': -1*v_dict['lcgr'],                          # Right rear wheel
                                  'rrw_a': -1*v_dict['lcgr'] + v_dict['tire_d']/2,
                                  'rrw_b': -1*v_dict['lcgr'] + v_dict['tire_d']/2,
                                  'rrw_c': -1*v_dict['lcgr'] - v_dict['tire_d']/2,
                                  'rrw_d': -1*v_dict['lcgr'] - v_dict['tire_d']/2,
                                  'lrw': -1*v_dict['lcgr'],                          # Left rear wheel
                                  'lrw_a': -1*v_dict['lcgr'] + v_dict['tire_d']/2,
                                  'lrw_b': -1*v_dict['lcgr'] + v_dict['tire_d']/2,
                                  'lrw_c': -1*v_dict['lcgr'] - v_dict['tire_d']/2,
                                  'lrw_d': -1*v_dict['lcgr'] - v_dict['tire_d']/2,
                                  'cg': 0,                                         # CG
                                  'xaxis':v_dict['lcgf'] + v_dict['f_hang']+1.5,   # line for x-axis
                                  'yaxis': 0,                                      # line for y-axis
                                  'vel_v':10*math.cos(vin.beta_rad[i])}, ignore_index=True)  # line for velocity vector

        draw_vy = draw_vy.append({'b_lfc': -1*v_dict['width']/2,     # body outline
                                  'b_rfc': v_dict['width']/2,
                                  'b_rrc': v_dict['width']/2,
                                  'b_lrc': -1*v_dict['width']/2,
                                  'lfw':  -1 * (v_dict['width']/2 - v_dict['tire_w']/2),                         # left front wheel
                                  'lfw_a': -1 * (v_dict['width']/2 - v_dict['tire_w']/2) + v_dict['tire_d']/2*math.sin(vin.loc[i,'delta_rad']) + -1*v_dict['tire_w']/2*math.cos(vin.loc[i,'delta_rad']),
                                  'lfw_b': -1 * (v_dict['width']/2 - v_dict['tire_w']/2) + v_dict['tire_d']/2*math.sin(vin.loc[i,'delta_rad']) + v_dict['tire_w']/2*math.cos(vin.loc[i,'delta_rad']),
                                  'lfw_c': -1 * (v_dict['width']/2 - v_dict['tire_w']/2) + -1*v_dict['tire_d']/2*math.sin(vin.loc[i,'delta_rad']) + v_dict['tire_w']/2*math.cos(vin.loc[i,'delta_rad']),
                                  'lfw_d': -1 * (v_dict['width']/2 - v_dict['tire_w']/2) + -1*v_dict['tire_d']/2*math.sin(vin.loc[i,'delta_rad']) + -1*v_dict['tire_w']/2*math.cos(vin.loc[i,'delta_rad']),
                                  'rfw': (v_dict['width']/2 - v_dict['tire_w']/2),                         # Right front wheel
                                  'rfw_a': (v_dict['width']/2 - v_dict['tire_w']/2) + v_dict['tire_d']/2*math.sin(vin.loc[i,'delta_rad']) + -1*v_dict['tire_w']/2*math.cos(vin.loc[i,'delta_rad']),
                                  'rfw_b':(v_dict['width']/2 - v_dict['tire_w']/2) + v_dict['tire_d']/2*math.sin(vin.loc[i,'delta_rad']) + v_dict['tire_w']/2*math.cos(vin.loc[i,'delta_rad']),
                                  'rfw_c':(v_dict['width']/2 - v_dict['tire_w']/2) + -1*v_dict['tire_d']/2*math.sin(vin.loc[i,'delta_rad']) + v_dict['tire_w']/2*math.cos(vin.loc[i,'delta_rad']),
                                  'rfw_d':(v_dict['width']/2 - v_dict['tire_w']/2) + -1*v_dict['tire_d']/2*math.sin(vin.loc[i,'delta_rad']) + -1*v_dict['tire_w']/2*math.cos(vin.loc[i,'delta_rad']),
                                  'rrw': (v_dict['width']/2 - v_dict['tire_w']/2),                          # Right rear wheel
                                  'rrw_a':v_dict['width']/2 - v_dict['tire_w'],
                                  'rrw_b': v_dict['width']/2,
                                  'rrw_c':v_dict['width']/2,
                                  'rrw_d':v_dict['width']/2 - v_dict['tire_w'],
                                  'lrw': -1 * (v_dict['width']/2 - v_dict['tire_w']/2),                          # Left rear wheel
                                  'lrw_a': -1 * v_dict['width']/2,
                                  'lrw_b': -1 * v_dict['width']/2 + v_dict['tire_w'],
                                  'lrw_c': -1 * v_dict['width']/2 + v_dict['tire_w'],
                                  'lrw_d': -1 * v_dict['width']/2,
                                  'cg': 0,                                         # CG
                                  'xaxis':0,   # line for x-axis
                                  'yaxis': v_dict['width']/2+1.5,                                      # line for y-axis
                                  'vel_v':10*math.sin(vin.beta_rad[i])}, ignore_index=True)  # line for velocity vector

        # translate vehicle coordinates above to global frame
        # create new dataframes for global coordinates

    cgx = vin[['Dx']].copy()                                                    # CG location in inertial frame
    cgy = vin[['Dy']].copy()

    draw_gx = draw_vx.copy()                                                    # time variant vehicle geometry for plotting motion in inertial frame
    draw_gy = draw_vy.copy()                                                    # time variant vehicle geometry for plotting motion in inertial frame

    for i in range(0, len(draw_vx)):                                            # loop for coordinate transformation
           #print(i)
           for j in range(0, len(draw_vx.columns)):
              #print(i,j)
              # coordinate transformation - rotating vehicle frame to draw vehicle
              draw_gx.iloc[i, j] = cgx.loc[i,'Dx'] + draw_vx.iloc[i,j]*math.cos(vin.loc[i,'theta_rad']) - draw_vy.iloc[i,j] * math.sin(vin.loc[i,'theta_rad'])
              draw_gy.iloc[i, j] = cgy.loc[i,'Dy'] + draw_vx.iloc[i,j]*math.sin(vin.loc[i,'theta_rad']) + draw_vy.iloc[i,j] * math.cos(vin.loc[i,'theta_rad'])
              #draw_gx.loc[i, 'vel_v'] = 10*math.cos(vin.beta_rad[i])
              #draw_gy.loc[i, 'vel_v'] = 8*math.sin(vin.beta_rad[i])

    # copy locked info from vehicle model
    draw_gx['lf_lock'] = vin.lf_lock.copy()
    draw_gx['rf_lock'] = vin.rf_lock.copy()
    draw_gx['rr_lock'] = vin.rr_lock.copy()
    draw_gx['lr_lock'] = vin.lr_lock.copy()

    return draw_vx, draw_vy, draw_gx, draw_gy



"""
functions for detecting impact using point (Veh1) and edge (Veh2)
"""
import pandas as pd
import numpy as np
import math

def detect(i, impactNum, vehicle_list, impPointEdge, strikingVehicle, struckVehicle, crush_data):
    """
    detects impacts in sequence using impact loaction and contant plane lists
    """
    #print(f'Vehicle 1 time = {vehicle_list[0].model.t[i]}')
    # current location of impact point - vehicle 1
    cgx1 = vehicle_list[strikingVehicle].model.Dx[i]  # take last value in Dx
    cgy1 = vehicle_list[strikingVehicle].model.Dy[i]  # take last value in Dy
    theta_rad1 = vehicle_list[strikingVehicle].model.theta_rad[i]

    impactp_gx = cgx1 + impPointEdge[impactNum]['impact_points'][0] * math.cos(theta_rad1) - impPointEdge[impactNum]['impact_points'][1] * math.sin(theta_rad1)
    impactp_gy = cgy1 + impPointEdge[impactNum]['impact_points'][0] * math.sin(theta_rad1) + impPointEdge[impactNum]['impact_points'][1] * math.cos(theta_rad1)

    # impact edge - vehicle 2
    cgx2 = vehicle_list[struckVehicle].model.Dx[i]
    cgy2 = vehicle_list[struckVehicle].model.Dy[i]
    theta_rad2 = vehicle_list[struckVehicle].model.theta_rad[i]

    # impact point in veh2 reference frame
    impactp_veh2x = ((impactp_gx-cgx2) * math.cos(theta_rad2) + (impactp_gy-cgy2) * math.sin(theta_rad2))
    impactp_veh2y = (-1 * (impactp_gx-cgx2) * math.sin(theta_rad2) + (impactp_gy-cgy2) * math.cos(theta_rad2))

    """
    calculate relative position of impact point and impact edge
    save data to refer to determine relative velocity
    edge_loc is the location of the impact point along the impact edge,
    moving clockwise
    """
    # edgeimpact_points = (x1, y1, x2, y2)

    crush_data.impact[i] = False
    if vehicle_list[struckVehicle].edgeimpact == 0:
        crush_data.normal_crush[i] = impactp_veh2x - impPointEdge[impactNum]['edgeimpact_points'][0]
        crush_data.edge_loc[i] = impPointEdge[impactNum]['edgeimpact_points'][1] - impactp_veh2y
        if (crush_data.normal_crush[i] <= 0) & (impactp_veh2y >= impPointEdge[impactNum]['edgeimpact_points'][1]) & (impactp_veh2y <= impPointEdge[impactNum]['edgeimpact_points'][3]):
            crush_data.impact[i] = True
    elif vehicle_list[struckVehicle].edgeimpact == 1:
        crush_data.normal_crush[i] = impactp_veh2y - impPointEdge[impactNum]['edgeimpact_points'][1]
        crush_data.edge_loc[i] = impPointEdge[impactNum]['edgeimpact_points'][0] - impactp_veh2x
        if (crush_data.normal_crush[i] <= 0) & (impactp_veh2x <= impPointEdge[impactNum]['edgeimpact_points'][0]) & (impactp_veh2x >= impPointEdge[impactNum]['edgeimpact_points'][2]):
            crush_data.impact[i] = True
    elif vehicle_list[struckVehicle].edgeimpact == 2:
        crush_data.normal_crush[i] = impPointEdge[impactNum]['edgeimpact_points'][0] - impactp_veh2x
        crush_data.edge_loc[i] = impPointEdge[impactNum]['edgeimpact_points'][1] - impactp_veh2y
        if (crush_data.normal_crush[i] <= 0) & (impactp_veh2y <= impPointEdge[impactNum]['edgeimpact_points'][1]) & (impactp_veh2y >= impPointEdge[impactNum]['edgeimpact_points'][3]):
            crush_data.impact[i] = True
    elif vehicle_list[struckVehicle].edgeimpact == 3:
        crush_data.normal_crush[i] = impPointEdge[impactNum]['edgeimpact_points'][1] - impactp_veh2y
        crush_data.edge_loc[i] = impactp_veh2x - impPointEdge[impactNum]['edgeimpact_points'][0]
        if (crush_data.normal_crush[i] < 0) & (impactp_veh2x >= impPointEdge[impactNum]['edgeimpact_points'][0]) & (impactp_veh2x <= impPointEdge[impactNum]['edgeimpact_points'][2]):
            crush_data.impact[i] = True

    crush_data.ImpactNumber[i] = impactNum
    crush_data.impactp_veh2x[i] = impactp_veh2x
    crush_data.impactp_veh2y[i] = impactp_veh2y

    return crush_data



import pandas as pd
import numpy as np
from copy import deepcopy
from .impact_vehicle_model import multi_vehicle_model
from .model_calcs.impact_detect import detect
from .model_calcs.collision_plane import define_impact_plane, define_impact_edge
from pycrash.model_calcs.position_data import position_data_motion, position_data_static
from .visualization.model_interval_two_vehicles import plot_motion_interval
from .visualization.initial_positions import initial_position
from .model_calcs.carpenter_momentum_calcs import impc
from .visualization.kinematics import plot_model
from .visualization.vehicles_at_impact import plot_impact
# TODO: add ability to use force-displacement data
# TODO: change self.sim_defaults to sim_settings
# TODO: environment data

# column list for vehicle model
vehicle_data_columns = ['t', 'vx', 'vy', 'Vx', 'Vy', 'Vr', 'vehicleslip_deg', 'vehicleslip_rad', 'oz_deg', 'oz_rad', 'delta_deg',
                        'delta_rad', 'turn_rX', 'turn_rY', 'turn_rR', 'au', 'av', 'ax', 'ay', 'ar', 'Ax', 'Ay', 'Ar',
                        'alphaz', 'alphaz_deg', 'beta_deg', 'beta_rad', 'lf_fx', 'lf_fy', 'rf_fx', 'rf_fy',
                        'rr_fx', 'rr_fy', 'lr_fx', 'lr_fy', 'lf_alpha', 'rf_alpha', 'rr_alpha', 'lr_alpha',
                        'lf_lock', 'rf_lock', 'rr_lock', 'lr_lock', 'lf_fz', 'rf_fz', 'rr_fz', 'lr_fz',
                        'theta_rad', 'theta_deg', 'Fx', 'Fy', 'Mz']

#model = pd.DataFrame(np.zeros(shape=(1, len(vehicle_data_columns))), columns=vehicle_data_columns)

""" create inputs for impact """
def create_impact_order():
    """ create list of impacts with list of striking and struck vehicle for each impact """
    print('Creating list of striking and struck vehicle for each impact')
    print('Vehicle indices start at 0')
    impactTotal = int(input("Enter total number of impacts: "))
    impact_order = []
    for i in range(0, impactTotal):
        striking = int(input(f"Enter striking vehicle number for impact {i}: "))
        struck = int(input(f"Enter struck vehicle number for impact {i}: "))
        impact_order.append([striking, struck])

    print(f'Impact order defined as: {impact_order}')
    return impact_order

def create_impc_inputs(numImpacts):
    """ create dictionary for each impact with intervehicle friction and restitution
        defined for each impact  """
    impc_inputs = {}
    for i in range(0, numImpacts):
        mu = float(input(f"Enter intervehicular friction for impact {i}: "))
        cor = float(input(f"Enter coefficient of restitution for impact {i}: "))
        impc_inputs[i] = {'vehicle_mu': mu,
                          'cor': cor}

    print(f'IMPC inputs defined as: {impc_inputs}')
    return impc_inputs



"""
main file for controlling vehicle motion simulation and impact
"""

class Impact():
    def __init__(self, name, endTime, impact_type, vehicle_list, impact_order=None, impc_inputs=None, user_sim_defaults=None):
        """ impact_order defines the [striking , struck] vehicle using a list of lists
        impc_inputs is a dictionary of inputs for each impact of the form {0:{'vehicle_mu': 0.3, 'cor': 0.1}}"""
        self.name = name
        self.type = 'impact'
        self.vehicles = deepcopy(vehicle_list)
        self.impactNum = 0  # counter for the number of impacts in simulation
        self.endTime = endTime  # total simulation time (s)
        self.impc_inputs = impc_inputs  # dictionary of inputs for IMPC by impact

        # initialize vehicle variables
        for veh in self.vehicles:
            veh.impactTotal = 0

        """ test for impact_order """
        if impact_order:
            print(f'Impact order defined as: {impact_order}')
            self.impact_order = impact_order
        else:
            self.impact_order = create_impact_order()

        """ get number of impacts per vehicle """
        self._impactsPerVehicle = [0] * len(self.vehicles)  # count impacts per vehicle
        for imp in self.impact_order:
            self._impactsPerVehicle[imp[0]] += 1
            self._impactsPerVehicle[imp[1]] += 1
            self.vehicles[imp[0]].impactTotal += 1
            self.vehicles[imp[1]].impactTotal += 1

        """ test for impact_inputs """
        if impc_inputs:
            print(f'IMPC inputs defined as: {impc_inputs}')
            if len(impc_inputs) != len(impact_order):
                print(f'{len(impc_inputs)} IMPC inputs provided. Impact order indicates {len(impact_order)} are required.')
            else:
                self.impc_inputs = impc_inputs
        else:
            self.impc_inputs = create_impc_inputs(len(self.impact_order))

        if (impact_type not in ["SS", "IMPC", "impc"]):
            print("Not a valid impact type, choose SS, IMPC")
            impact_type = input("Enter an impact type of SS or IMPC: ")

            if (impact_type not in ["SS", "IMPC", "impc"]):
                print("Not a valid impact type - value set to None")
                self.impact_type == None
            else:
                self.impact_type = impact_type
        else:
            self.impact_type = impact_type   # <- assign impact type to provided value

        # load defaults
        if user_sim_defaults:
            # TODO: create check for user sim_defaults_input
            self.sim_defaults = user_sim_defaults
        else:
            self.sim_defaults = {'dt_motion': 0.01,
                                 'mu_max': 0.76,
                                 'alpha_max': 0.174533,  # 10 degrees
                                }

        # check for driver inputs
        for veh in self.vehicles:
            if hasattr(veh, 'driver_input'):
                print(f"Driver input for {veh.name} of shape = {veh.driver_input.shape}")
            else:
                print(f'Driver input for {veh.name} not provided - no braking or steering applied')
                end_time = int(input('Enter duration for simulation (seconds):'))
                t = list(np.arange(0, end_time + self.sim_defaults['dt_motion'], self.sim_defaults['dt_motion']))  # create time array from 0 to end time from user
                throttle = [0] * len(t)
                brake = [0] * len(t)
                steer = [0] * len(t)
                veh.driver_input = pd.DataFrame.from_dict({'t': t, 'throttle': throttle, 'brake': brake, 'steer': steer})
                print(f'Driver inputs for {veh.name} set to zero for {end_time} seconds')

        # TODO: convert sideswipe to lists of tuples like impc
            # sideswipe will not have iplane=True
        """ IMPC model - impact points and edge detection """
        # TODO: test for totalIMpacts = number of impact points / edges
        if self.impact_type in ['IMPC', 'impc']:
            for veh in self.vehicles:
                if veh.striking:
                    if hasattr(veh, 'impact_points'):
                        print(f'Predefined impact points for {veh.name}: [x (ft), y (ft), impact plane angle (deg)] = {veh.impact_points}')
                        veh._impactPoints = len(veh.impact_points)
                    else:
                        print(f"Create impact point(s) for {veh.name} [striking vehicle]")
                        print("")
                        veh = define_impact_plane(veh, iplane=True)
                    print(f'Total impacts for {veh.name}: {len(veh.impact_points)}')

                else:
                    if hasattr(veh, 'edgeimpact_points'):
                        print(f'Predefined impact edge for {veh.name}: {veh.edgeimpact}')
                    else:
                        print(f"Create impacting edge for {veh.name} [stuck vehicle]")
                        print("")
                        veh = define_impact_edge(veh, iplane=False)
                    print(f'Total impacts for {veh.name} - total defined edges: {len(veh.edgeimpact_points)}')

        """ initialize vehicle motion dataframes """
        # TODO: resolve differences between vehicle input time duration and simulation duration - input must be > than simulation?

        for veh in self.vehicles:
            if veh.type == 'Barrier':
                # create vehicle with no motion for "Barrier" type
                veh.model = pd.DataFrame(np.nan, index=np.arange(1+len(np.arange(0, self.endTime, self.sim_defaults['dt_motion']))), columns=vehicle_data_columns)
                veh.model['Dx'] = veh.init_x_pos
                veh.model['Dy'] = veh.init_y_pos
                veh.model = veh.model.fillna(0)
            else:
                veh.model = pd.DataFrame(np.nan, index=np.arange(1+len(np.arange(0, self.endTime, self.sim_defaults['dt_motion']))), columns=vehicle_data_columns)
                veh.model.Fx[0] = 0
                veh.model.Fy[0] = 0
                veh.model.Mz[0] = 0
                veh.model.au[0] = 0  # no initial vehicle pitch
                veh.model.av[0] = 0  # no initial vehicle roll
                veh.model.vx[0] = veh.vx_initial * 1.46667  # convert input in mph to fps
                veh.model.vy[0] = veh.vy_initial * 1.46667  # convert input in mph to fps
                veh.model.theta_rad[0] = veh.head_angle * np.pi / 180  # initial heading angle
                veh.model.oz_rad[0] = veh.omega_z * (np.pi / 180)  # initial angular rate (deg/s) - input
                veh.model.Vx[0] = veh.vx_initial * 1.46667 * np.cos(veh.head_angle * np.pi / 180) - veh.vy_initial * 1.46667 * np.sin(veh.head_angle * np.pi / 180)
                veh.model.Vy[0] = veh.vx_initial * 1.46667 * np.sin(veh.head_angle * np.pi / 180) + veh.vy_initial * 1.46667 * np.cos(veh.head_angle * np.pi / 180)

        self.detect_data = pd.DataFrame(np.nan, index=np.arange(1+len(np.arange(0, self.endTime, self.sim_defaults['dt_motion']))),
                                   columns=['impact', 'edge_loc', 'normal_crush', 'impactp_veh2x', 'impactp_veh2y', 'ImpactNumber'])

        """ dictionaries for storing impact related data """
        self._impactList = list(range(0, len(self.impact_order)))
        self.impc_results = dict.fromkeys(self._impactList)
        self._impPointEdge = dict.fromkeys(self._impactList)
        self._impactIndex = {}  # index of each impact for plotting

        # create list of striking and struck vehicle
        self._strikingVehicleList = []
        self._struckVehicleList = []
        impactsPerVehicle = [0] * len(self.vehicles)  # count impacts per vehicle
        counter = 0
        for imp in self.impact_order:
            self._strikingVehicleList.append(imp[0])  # striking vehicle in each impact
            self._struckVehicleList.append(imp[1])    # struck vehicle in each impact

            # dictionary of impact points and edges for each impact - starts at zero
            self._impPointEdge[counter] = {'impact_points': self.vehicles[imp[0]].impact_points[impactsPerVehicle[imp[0]]],
                                           'edgeimpact_points': self.vehicles[imp[1]].edgeimpact_points[impactsPerVehicle[imp[1]]]}  # impact points and edge for each impact
            impactsPerVehicle[imp[0]] += 1
            impactsPerVehicle[imp[1]] += 1
            counter += 1

        print(self._impPointEdge)
        """ create impact object is complete """
        print(f'Impact simulation created with {len(self.vehicles)} vehicles of type {self.impact_type}')
        print(f'Simulation inputs: {self.sim_defaults}')

    """ plot initial positions: include impact planes """
    # plot initial position

    def show_initial_position(self):
        initial_position(position_data_static(self.vehicles))

    """ plot vehicle motion """
    def plot_vehicle_motion(self, n_intervals, tire_path=True, show_vector=False):
        # TODO: adjust what impact point to plot
        plotImpactPoint = 0     # starts at 1
        plot_motion_interval(self.vehicles, self._impactIndex, n_intervals,
                             tire_path=tire_path, show_vector=show_vector)

    """ plot vehicles at impact """
    def plot_impact(self, impactNum):
            for key, value in self._impactIndex.items():
                if value == impactNum:
                    plot_impact(self.vehicles, key-1, self._impactIndex, show_vector=True)
                    print(f'Plotting impact {value} at index {key}')

    """ plot kinematic data """
    def plot_model(self, vehNum):
        plot_model(self.vehicles[vehNum])

    """ initiate impact simulation """

    def simulate(self, show_results=True):
        """
        end_time <- default time of simulation (seconds), simulation will still stop if vehicle motion is zero
        """
        self.show_results = show_results   # print results from impact simulation calculations
        separation = True   # vehicles are separated / not engaged
        impactsComplete = False  # all impacts have occurred

        for i in range(0, 1 + int(self.endTime / self.sim_defaults['dt_motion'])):
            #print(i)
            for veh in self.vehicles:
                """ calculate current position of each vehicle """
                if veh.type != 'Barrier':
                    #print(f'Simulating Vehicle: {veh.name}')
                    veh = multi_vehicle_model(veh, i, self.sim_defaults, self.impact_type)

            if not impactsComplete:

                """ use impact detect to look for impacts"""
                strikingVehicle = self.impact_order[self.impactNum][0]  # index of striking vehicle
                struckVehicle = self.impact_order[self.impactNum][1]    # index of struck vehicle

                # separate crush data for each impact - create dictionary for each impact
                #print(f'Looking for impact {self.impactNum} between vehicles: {self.impact_order[self.impactNum]}')
                #print(f'vehicles are separated: {separation}')
                self.detect_data = detect(i, self.impactNum, self.vehicles, self._impPointEdge, strikingVehicle, struckVehicle, self.detect_data)

                if self.detect_data.impact[i]:
                    print('')
                    print(f"Impact #{self.impactNum} detected at i: {i}, t: {i * self.sim_defaults['dt_motion']}")
                    print('')
                    """ apply momentum IMPC if impact only if vehicles are separated """
                    if (self.detect_data.impact[i] & separation):
                        if self.impact_type in ["IMPC", "impc"]:

                            self.vehicles, self.impc_results[self.impactNum] = impc(i,
                                                                                   self.impactNum,
                                                                                   self.detect_data.impactp_veh2x[i],
                                                                                   self.detect_data.impactp_veh2y[i],
                                                                                   self._impPointEdge,
                                                                                   self.vehicles,
                                                                                   strikingVehicle,
                                                                                   struckVehicle,
                                                                                   self.impc_inputs[self.impactNum],
                                                                                   self.sim_defaults['dt_motion'],
                                                                                   self.show_results)  # run impc model

                        separation = False
                        self._impactIndex[i] = self.impactNum
                        self.impactNum += 1

                else:
                    if (not separation) and (not self.detect_data.impact[i]):
                        # separation is false and impact is true - vehicles have now separated
                        print(f'Vehicles separated')
                        separation = True

            if self.impactNum == len(self.impact_order):
                impactsComplete = True

        """ convert and position data for plots """
        for veh in self.vehicles:
            veh.model.alphaz_deg = [row * 180 / np.pi for row in veh.model.alphaz]  # move to seperate calc
            veh.model.oz_deg = [row * 180 / np.pi for row in veh.model.oz_rad]  # move to seperate calc
            veh.model.theta_deg = [row * 180 / np.pi for row in veh.model.theta_rad]  # move to seperate calc
            veh.model.beta_deg = [row * 180 / np.pi for row in veh.model.beta_rad]  # move to seperate calc
            # get position data
            veh = position_data_motion(veh)








import plotly.graph_objects as go
import math

width = 1000
aspect_ratio = 16 / 9
figure_size = (width, width / aspect_ratio)
wheel_colors = ['rgb(0, 0, 255)', 'rgb(0, 255, 0)', 'rgb(153, 0, 204)', 'rgb(255, 102, 0)']
# plot initial positions and any motion data to show vehicle paths
def initial_position(veh1, veh2, i):
    # plot initial positions
    # grid grid based on initial position of vehicle
    # scale x,y axes accordingly
    # static geometry from model_calcs.position_data

    """
    combine two vehicles into a single df, to allow for max / min check
    """
    fig = go.Figure()
    for veh in [veh1, veh2]:
        # Body and tire outlines
        bdy_x = (veh.Px.b_lfc[i], veh.Px.b_rfc[i], veh.Px.b_rrc[i], veh.Px.b_lrc[i], veh.Px.b_lfc[i])
        bdy_y = (veh.Py.b_lfc[i], veh.Py.b_rfc[i], veh.Py.b_rrc[i], veh.Py.b_lrc[i], veh.Py.b_lfc[i])

        lfw_x = (veh.Px.lfw_a[i], veh.Px.lfw_b[i], veh.Px.lfw_c[i], veh.Px.lfw_d[i], veh.Px.lfw_a[i])
        lfw_y = (veh.Py.lfw_a[i], veh.Py.lfw_b[i], veh.Py.lfw_c[i], veh.Py.lfw_d[i], veh.Py.lfw_a[i])

        rfw_x = (veh.Px.rfw_a[i], veh.Px.rfw_b[i], veh.Px.rfw_c[i], veh.Px.rfw_d[i], veh.Px.rfw_a[i])
        rfw_y = (veh.Py.rfw_a[i], veh.Py.rfw_b[i], veh.Py.rfw_c[i], veh.Py.rfw_d[i], veh.Py.rfw_a[i])

        rrw_x = (veh.Px.rrw_a[i], veh.Px.rrw_b[i], veh.Px.rrw_c[i], veh.Px.rrw_d[i], veh.Px.rrw_a[i])
        rrw_y = (veh.Py.rrw_a[i], veh.Py.rrw_b[i], veh.Py.rrw_c[i], veh.Py.rrw_d[i], veh.Py.rrw_a[i])

        lrw_x = (veh.Px.lrw_a[i], veh.Px.lrw_b[i], veh.Px.lrw_c[i], veh.Px.lrw_d[i], veh.Px.lrw_a[i])
        lrw_y = (veh.Py.lrw_a[i], veh.Py.lrw_b[i], veh.Py.lrw_c[i], veh.Py.lrw_d[i], veh.Py.lrw_a[i])


        # body outline
        fig.add_trace(go.Scatter(x = bdy_x, y = bdy_y,
                                    mode = 'lines',
                                    name = f'{veh.name} - body',
                                    line = dict(color = 'rgb(0, 0, 0)', width = 2)))

        # wheel outline
        fig.add_trace(go.Scatter(x = lfw_x, y = lfw_y,
                            mode = 'lines',
                            name = 'LF',
                            line = dict(color = wheel_colors[0], width = 1)))
        fig.add_trace(go.Scatter(x = rfw_x, y = rfw_y,
                            mode = 'lines',
                            name = 'RF',
                            line = dict(color = wheel_colors[1], width = 1)))
        fig.add_trace(go.Scatter(x = rrw_x, y = rrw_y,
                            mode = 'lines',
                            name = 'RR',
                            line = dict(color = wheel_colors[2], width = 1)))
        fig.add_trace(go.Scatter(x = lrw_x, y = lrw_y,
                            mode = 'lines',
                            name = 'LR',
                            line = dict(color = wheel_colors[3], width = 1)))

        # CG
        cgx = [veh.Px.cg[i], veh.Px.cg[i]]
        cgy = [veh.Py.cg[i], veh.Py.cg[i]]

        # adjust axes to keep aspect aspect ratio
        """
        dx_max = veh.p_gx.cg.max() + 15
        dx_min = veh.p_gx.cg.min() - 15
        dy_max = veh.p_gy.cg.max() + 15
        dy_min = veh.p_gy.cg.min() - 15

        dx = dx_max - dx_min
        dy = dy_max - dy_min

        if dx > dy:
            adj_x = aspect_ratio * dy / dx
            adj_y = 1
            print(f" dx > dy -> adj_x = {adj_x}, adj_y = {adj_y}")
            dx_min = round(dx_min * adj_x)
            dx_max = round(dx_max * adj_x)
            print(f"dx_min = {dx_min}, dx_max = {dx_max}")
            print(f"dy_min = {dy_min}, dy_max = {dy_max}")
        else:
            adj_x = 1
            adj_y = (1 / aspect_ratio) * dx / dy
            print(f" dy > dx -> adj_x = {adj_x}, adj_y = {adj_y}")
            dy_min = round(dy_min * adj_y)
            dy_max = round(dy_max * adj_y)
            print(f"dx_min = {dx_min}, dx_max = {dx_max}")
            print(f"dy_min = {dy_min}, dy_max = {dy_max}")
        """

        fig.add_trace(go.Scatter(x = cgx, y = cgy,
                            mode = 'markers',
                            name = f'CG - {veh.name}',
                            marker = dict(color = 'rgb(0, 0, 0)', size = 10),
                            ))


        # velocity vector
        fig.add_annotation(x = veh.Px.vel_v[i],
                           y = veh.Py.vel_v[i],
                           ax = veh.Px.cg[i],
                           ay = veh.Py.cg[i],
                           axref = 'x',
                           ayref = 'y',
                           xref = 'x',
                           yref = 'y',
                           text = "",
                           showarrow = True,
                           arrowsize = 2.5,
                           arrowhead = 1,
                           arrowwidth = 1.7,
                           arrowcolor = 'rgb(255, 0, 0)')

        # x axis
        fig.add_annotation(x = veh.Px.xaxis[i],
                           y = veh.Py.xaxis[i],
                           ax = veh.Px.cg[i],
                           ay = veh.Py.cg[i],
                           axref = 'x',
                           ayref = 'y',
                           xref = 'x',
                           yref = 'y',
                           text = "",
                           showarrow = True,
                           arrowsize = 2,
                           arrowhead = 1,
                           arrowwidth = 1.5,
                           arrowcolor = 'rgb(0, 0, 0)')

        # y axis
        fig.add_annotation(x = veh.Px.yaxis[i],
                           y = veh.Py.yaxis[i],
                           ax = veh.Px.cg[i],
                           ay = veh.Py.cg[i],
                           axref = 'x',
                           ayref = 'y',
                           xref = 'x',
                           yref = 'y',
                           text = "",
                           showarrow = True,
                           arrowsize = 2,
                           arrowhead = 1,
                           arrowwidth = 1.5,
                           arrowcolor = 'rgb(0, 0, 255)')

    fig.update_layout(
        showlegend = False,
        autosize = False,
        width = width,
        height = width / aspect_ratio,
        title = 'Initial Vehicle Positions',
        template = 'plotly_white',
        xaxis = dict(showgrid = True, title = 'Forward (ft)'),
        yaxis = dict(showgrid = True, title = 'Rightward (ft)'),
        font = dict(family = 'Arial', size = 16, color = 'black'))

    fig.update_xaxes(showline=True, linewidth=1, linecolor='black', ticks="outside",
                     tickwidth=1, tickcolor='black', ticklen=10, zeroline=False)
    fig.update_yaxes(autorange="reversed", showline=True, linewidth=1, linecolor='black', ticks="outside",
                     tickwidth=1, tickcolor='black', ticklen=10, zeroline=False)

    fig.show()

    # Table of initial conditions

    fig = go.Figure(data=[go.Table(
    header=dict(values=['Vehicle', 'Vx (mph)', 'Vy (mph)', 'V (mph)', 'Omega (deg/s)', 'X Position (ft)', 'Y Position (ft)'],
                line_color='darkslategray',
                fill_color='lightskyblue',
                align='left'),
    cells=dict(values=[[f'{veh1.name}', f'{veh2.name}'],
                       [veh1.vx_initial, veh2.vx_initial],
                       [veh1.vy_initial, veh2.vy_initial],
                       [math.sqrt(veh1.vx_initial**2 + veh1.vy_initial**2), math.sqrt(veh2.vx_initial**2 + veh2.vy_initial**2)],
                       [veh1.omega_z, veh2.omega_z],
                       [veh1.init_x_pos, veh2.init_x_pos],
                       [veh1.init_y_pos, veh2.init_y_pos]],
               line_color='darkslategray',
               fill_color='lightcyan',
               align='left'))
               ])

    fig.update_layout(width=500, height=300)
    fig.show()



"""
create a collision plane on the striking vehicle
this will indicate the contact point and direction of the nornmal and tangential axis of impact force / momentum balance
"""
from pycrash.visualization.vehicle import plot_impact_points, plot_impact_edge
import math

def define_impact_plane(veh, iplane=True):
    """
    generates a point in vehicle 1 (striking) reference frame
    sideswipe collisions - px, py will determine the extent of vehcle engagement with respect
    to contacting edge in vehicle 2 (struck)
    impact momentum model - px, py will be used along with the impact plane to determine time
    of impact and direction of normal and tangential contact planes
    when plane = True, user will also indicate direction of normal impact plane
    """
    user_loc = False
    # set impact edge to zero
    veh.striking = True

    # test for required inputs
    if not veh.lcgf:
        veh.lcgf = float(input("Enter CG to front axle (ft)"))

    if not veh.lcgr:
        veh.lcgr = float(input("Enter CG to rear axle (ft)"))

    if not veh.f_hang:
        veh.f_hang = float(input("Enter front overhang (ft)"))

    if not veh.r_hang:
        veh.r_hang = float(input("Enter rear overhang (ft)"))

    if not veh.width:
        veh.width = float(input("Enter vehicle width (ft)"))

    # create figure of vehicle 1 with scale / grid and p1, p2, p3, p4 labeled when function is called
    # option 5 = custom location
    plot_impact_points(veh, iplane=False)  # plot vehicle points

    veh.impact_points = []  # list of impact points x, y and normal impact plane angle deg
    veh.impact_norm = []    # list of points for plotting the impact plane
    veh.impact_tang = []    # list of points for plotting the impact plane

    """ create list of impact points - this will be a list of tuples"""
    for i in range(0, veh.impactTotal):
        impact_option = int(input("Choose option for impact point (1, 2, 3, 4, custom = 99: "))

        if impact_option not in [1, 2, 3, 4, 99]:
            print("Invalid impact point option - enter 1, 2, 3, 4 or 5")
            impact_option = int(input("Choose option for impact location"))
        elif impact_option != 99:
            if impact_option == 1:
                veh.impact_points.append((veh.lcgf + veh.f_hang, -1 * veh.width / 2))
            elif impact_option == 2:
                veh.impact_points.append((veh.lcgf + veh.f_hang, veh.width / 2))
            elif impact_option == 3:
                veh.impact_points.append((-1 * veh.lcgr - veh.r_hang, veh.width / 2))
            elif impact_option == 4:
                veh.impact_points.append((-1 * veh.lcgr - veh.r_hang, -1 * veh.width / 2))
        elif impact_option == 99:
            veh.impact_points.append((float(input("Enter x-coordinate ( + forward) of impact point in vehicle frame (ft):")),
                                      float(input("Enter y-coordinate ( + rightward) of impact point in vehicle frame (ft):"))))
            user_loc = True

        if iplane:
            veh.impact_points[i] += (float(input("Enter heading angle of normal impact plane (deg):  ")), )  # append to x,y cooridinates
            norm_length = veh.width
            tang_length = veh.width / 2

            veh.impact_norm.append((veh.impact_points[i][0] + norm_length * math.cos(veh.impact_points[i][2] * math.pi / 180),
                                  veh.impact_points[i][1] + norm_length * math.sin(veh.impact_points[i][2] * math.pi / 180)))

            veh.impact_tang.append((veh.impact_points[i][0] + tang_length * math.cos(veh.impact_points[i][2] * math.pi / 180 + math.pi / 2),
                                  veh.impact_points[i][1] + tang_length * math.sin(veh.impact_points[i][2] * math.pi / 180 + math.pi / 2)))

        print(veh.impact_points)
        plot_impact_points(veh)  # plot vehicle points

    return veh

def define_impact_edge(veh, iplane = False):
    """
    generates an edge in vehicle 2 (struck) reference frame
    sideswipe collisions - edge will determine the extent of vehcle engagement with respect
    to contacting point in vehicle 1 (striking)
    impact momentum model - impact edge will be used along with the impact plane to determine time
    of impact
    """

    veh.striking = False

    # test for required inputs
    if not veh.lcgf:
        veh.lcgf = float(input("Enter CG to front axle (ft)"))

    if not veh.lcgr:
        veh.lcgr = float(input("Enter CG to rear axle (ft)"))

    if not veh.f_hang:
        veh.f_hang = float(input("Enter front overhang (ft)"))

    if not veh.r_hang:
        veh.r_hang = float(input("Enter rear overhang (ft)"))

    if not veh.width:
        veh.width = float(input("Enter vehicle width (ft)"))

    # create figure of vehicle 1 with scale / grid and p1, p2, p3, p4 labeled when function is called
    # option 5 = custom location


    veh.edgeimpact_points = []   # store x,y values for edge points - moving clockwise
    veh.edgeimpact = []    # defines location of edge

    plot_impact_edge(veh)


    for i in range(0, veh.impactTotal):
        impact_option = int(input("Choose impact edge for the corresponding impact point on striking vehicle: "))

        if impact_option not in [0, 1, 2, 3]:
            raise ValueError("Invalid impact edge option - enter 0, 1, 2, 3")
        else:
            if impact_option == 0:
                veh.edgeimpact.append(0)
                veh.edgeimpact_points.append((veh.lcgf + veh.f_hang, -1 * veh.width / 2,
                                              veh.lcgf + veh.f_hang, veh.width / 2))
            elif impact_option == 1:
                veh.edgeimpact.append(1)
                veh.edgeimpact_points.append((veh.lcgf + veh.f_hang, veh.width / 2,
                                              -1 * veh.lcgr - veh.r_hang, veh.width / 2))
            elif impact_option == 2:
                veh.edgeimpact.append(2)
                veh.edgeimpact_points.append((-1 * veh.lcgr - veh.r_hang, veh.width / 2,
                                              -1 * veh.lcgr - veh.r_hang, -1 * veh.width / 2))
            elif impact_option == 3:
                veh.edgeimpact.append(3)
                veh.edgeimpact_points.append((-1 * veh.lcgr - veh.r_hang, -1 * veh.width / 2,
                                              veh.lcgf + veh.f_hang, -1 * veh.width / 2))

    return veh



import pycrash.vehicle
import dash
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output

external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']

app = dash.Dash(__name__, external_stylesheets=external_stylesheets)

app.layout = html.Div([
    html.H6("Change the value in the text box to see callbacks in action!"),
    html.Div(["Input: ",
              dcc.Input(id='my-input', value='initial value', type='text')]),
    html.Br(),
    html.Div(id='my-output'),
])

@app.callback(
    Output(component_id='my-output', component_property='children'),
    Input(component_id='my-input', component_property='value')
)

def update_output_div(input_value):
    return 'Output: {}'.format(input_value)


if __name__ == '__main__':
    app.run_server(debug=True)



# -*- coding: utf-8 -*-
"""
Created on Mon May 13 21:58:23 2019

@author: JCormier
"""
# Contains all functions used to plot various elements of the program

import matplotlib.pyplot as plt
import numpy as np
import math

def plot_inputs(df_in):
    fig, axs = plt.subplots(2, 2, figsize=(18,10), sharey=False)
    axs[0,0].set_xlabel('time (s)')
    axs[0,0].set_title('EDR Speed, Angular Rate, Derived Acceleration', color='k')
    axs[0,0].plot(df_in.t, df_in.v_edr / 1.46667, color='k')
    axs[0,0].plot(df_in.t, df_in.oz_edr, color='g')
    axs[0,0].tick_params(axis='y', labelcolor='k')
    axs[0,0].set_ylabel('Velocity (mph) | Angular Rate (deg/s)', color='k')

    ax2 = []
    ax2.append(axs[0,0].twinx())
    ax2[0].set_ylabel('Derived Acceleration (g)', color='r')
    ax2[0].plot(df_in.t, df_in.a_edr/32.2, color='r')
    ax2[0].tick_params(axis='y', labelcolor='r')

    axs[0,1].set_xlabel('time (s)')
    axs[0,1].set_title('EDR Steering, Brake and Throttle', color='k')
    axs[0,1].tick_params(axis='y', labelcolor='k')
    axs[0,1].plot(df_in.t, df_in.sw_angle_edr, color='k')
    axs[0,1].plot(df_in.t, df_in.throttle_edr, color='r')
    axs[0,1].set_ylabel('Steering Wheel Angle (deg) | Throttle (%)', color='k')

    ax2.append(axs[0,1].twinx())
    ax2[1].plot(df_in.t, df_in.brake_edr, color='b')
    ax2[1].tick_params(axis='y', labelcolor='b')
    ax2[1].set_ylabel('Brake Status', color='b')

    axs[1,0].set_xlabel('time (s)')
    axs[1,0].set_title('Calculated Velocity (mph), Angular Rate (deg/s)', color='k')
    ax2.append(axs[1,0].twinx())
    ax2[2].set_ylabel('Acceleration (g)', color='r')

    axs[1,1].set_xlabel('time (s)')
    axs[1,1].set_title('Model Inputs: Steering Wheel Angle (deg), Input Throttle (%)', color='k')
    axs[1,1].tick_params(axis='y', labelcolor='k')
    axs[1,1].plot(df_in.t, df_in.sw_angle, color='k')
    axs[1,1].plot(df_in.t, df_in.throttle, color='r')
    axs[1,1].set_ylabel('Velocity (mph) | Steering Angle (deg)', color='k')

    ax2.append(axs[1,1].twinx())
    ax2[3].plot(df_in.t, df_in.brake, color='b')
    ax2[3].tick_params(axis='y', labelcolor='b')
    ax2[3].set_ylabel('Input Brake', color='b')

    plt.subplots_adjust(wspace=.6, hspace = .6)
    plt.show()

def plot_vehicle_pre_motion(df_in, pre_df):
    pre_df['V'] = np.sqrt(pre_df.Vx**2 + pre_df.Vy**2)/ 1.46667 
    fig, axs = plt.subplots(3, 2, figsize=(18,8), sharex='col')
    axs[0,0].set_xlabel('Time (s)')
    axs[0,0].set_ylabel('Velocity (mph) (Vx, Vy)', color='k')
    axs[0,0].plot(df_in.t - df_in.t[0], df_in.v_edr / 1.46667, color='k', linestyle='dashed', label = 'V - EDR')
    axs[0,0].plot(pre_df.t, pre_df.Vx / 1.46667, color='k', label = 'Vx')    
    axs[0,0].plot(pre_df.t, pre_df.Vy / 1.46667, color='b', label = 'Vy')
    axs[0,0].plot(pre_df.t, pre_df.V, color = 'g', linestyle='dashed', label = 'V')
    axs[0,0].tick_params(axis='y', labelcolor='k')
    axs[0,0].legend(frameon = False)

    #axs[0,1].set_xlabel('Time (s)')
    axs[0,1].set_ylabel('Acceleration (g)', color='k')
    axs[0,1].tick_params(axis='y', labelcolor='k')
    axs[0,1].plot(pre_df.t, pre_df.Ax/32.2, color='k', label = 'Ax')
    axs[0,1].plot(pre_df.t, pre_df.Ay/32.2, color='b', label = 'Ay')
    axs[0,1].plot(pre_df.t, pre_df.ax/32.2, color='k', linestyle='dashed', label = 'ax')
    axs[0,1].plot(pre_df.t, pre_df.ay/32.2, color='b', linestyle='dashed', label = 'ay')
    axs[0,1].legend(frameon = False)

    #axs[1,0].set_xlabel('Time (s)')
    axs[1,0].set_ylabel('Forward Tire Forces (lb)', color='k')
    axs[1,0].tick_params(axis='y', labelcolor='k')
    axs[1,0].plot(pre_df.t, pre_df.lf_fx, color='b')
    axs[1,0].plot(pre_df.t, pre_df.rf_fx, color='g')
    axs[1,0].plot(pre_df.t, pre_df.rr_fx, color='m')
    axs[1,0].plot(pre_df.t, pre_df.lr_fx, color='orange')

    #axs[1,1].set_xlabel('Time (s)')
    axs[1,1].set_ylabel('Rightward Tire Forces (lb)', color='k')
    axs[1,1].tick_params(axis='y', labelcolor='k')
    axs[1,1].plot(pre_df.t, pre_df.lf_fy, color='b')
    axs[1,1].plot(pre_df.t, pre_df.rf_fy, color='g')
    axs[1,1].plot(pre_df.t, pre_df.rr_fy, color='m')
    axs[1,1].plot(pre_df.t, pre_df.lr_fy, color='orange')

    axs[2,0].set_xlabel('Time (s)')
    axs[2,0].set_ylabel('Omega (deg/s), Alpha (deg/s/s)', color='k')
    axs[2,0].tick_params(axis='y', labelcolor='k')
    axs[2,0].plot(pre_df.t, pre_df.oz_deg, color='k')
    axs[2,0].plot(pre_df.t, pre_df.alphaz_deg, color='r')
    axs[2,0].plot(df_in.t - df_in.t[0], df_in.oz_edr, color='k', linestyle='dashed')

    axs[2,1].set_xlabel('Time (s)')
    axs[2,1].set_ylabel('Heading Angle (deg)', color='k')
    axs[2,1].tick_params(axis='y', labelcolor='k')
    axs[2,1].plot(pre_df.t, pre_df.theta_deg, color='k')

    #ax2 = []
    #ax2.append(axs[0].twinx())
    #ax2[0].set_ylabel('Acceleration (g) (ax, ay)', color='b')
    #ax2[0].plot(pre_df.t, pre_df.ax/32.2, color='b')
    #ax2[0].plot(pre_df.t, pre_df.ax/32.2, color='r')
    #ax2[0].tick_params(axis='y', labelcolor='b')

    #ax2.append(axs[1].twinx())
    #ax2[1].plot(df_in.t, df_in.brake_edr, color='b')
    #ax2[1].tick_params(axis='y', labelcolor='b')
    #ax2[1].set_ylabel('EDR Brake', color='b')

    plt.subplots_adjust(wspace=0.2, hspace = .1)
    plt.show()

def draw_vehicle(draw_vx, draw_vy, i):
    #  These dataframes link the components together for plotting
    bdy_x = (draw_vx.b_lfc[i], draw_vx.b_rfc[i], draw_vx.b_rrc[i], draw_vx.b_lrc[i], draw_vx.b_lfc[i])
    bdy_y = (draw_vy.b_lfc[i], draw_vy.b_rfc[i], draw_vy.b_rrc[i], draw_vy.b_lrc[i], draw_vy.b_lfc[i])

    lfw_x = (draw_vx.lfw_a[i], draw_vx.lfw_b[i], draw_vx.lfw_c[i], draw_vx.lfw_d[i], draw_vx.lfw_a[i])
    lfw_y = (draw_vy.lfw_a[i], draw_vy.lfw_b[i], draw_vy.lfw_c[i], draw_vy.lfw_d[i], draw_vy.lfw_a[i])

    rfw_x = (draw_vx.rfw_a[i], draw_vx.rfw_b[i], draw_vx.rfw_c[i], draw_vx.rfw_d[i], draw_vx.rfw_a[i])
    rfw_y = (draw_vy.rfw_a[i], draw_vy.rfw_b[i], draw_vy.rfw_c[i], draw_vy.rfw_d[i], draw_vy.rfw_a[i])

    rrw_x = (draw_vx.rrw_a[i], draw_vx.rrw_b[i], draw_vx.rrw_c[i], draw_vx.rrw_d[i], draw_vx.rrw_a[i])
    rrw_y = (draw_vy.rrw_a[i], draw_vy.rrw_b[i], draw_vy.rrw_c[i], draw_vy.rrw_d[i], draw_vy.rrw_a[i])

    lrw_x = (draw_vx.lrw_a[i], draw_vx.lrw_b[i], draw_vx.lrw_c[i], draw_vx.lrw_d[i], draw_vx.lrw_a[i])
    lrw_y = (draw_vy.lrw_a[i], draw_vy.lrw_b[i], draw_vy.lrw_c[i], draw_vy.lrw_d[i], draw_vy.lrw_a[i])

    #%% Plot Vehicle in Vehicle reference frame
    plt.figure(figsize=(16,9))
    plt.xlim([-20, 20])
    plt.ylim([-10, 10])

    plt.plot(bdy_x, bdy_y, 'k')
    plt.scatter(draw_vx.lfw[i], draw_vy.lfw[i], c='b')         # left front wheel center
    plt.plot(lfw_x, lfw_y, 'b')
    plt.scatter(draw_vx.rfw[i], draw_vy.rfw[i], c='g')       # right front wheel center
    plt.plot(rfw_x, rfw_y, 'g')
    plt.scatter(draw_vx.rrw[i], draw_vy.rrw[i],c='m')       # right rear wheel center
    plt.plot(rrw_x, rrw_y,'m')
    plt.scatter(draw_vx.lrw[i], draw_vy.lrw[i], c='orange')       # left rear wheel center
    plt.plot(lrw_x, lrw_y, 'orange')

    # vehicle CG
    plt.scatter(draw_vx.cg[i], draw_vy.cg[i],s=500, c='k')

    # velocity vector

    plt.arrow(draw_vx.cg[i], draw_vy.cg[i], draw_vx.vel_v[i]-draw_vx.cg[i], draw_vy.vel_v[i]-draw_vy.cg[i], head_width=.5, head_length=0.5, fc='r', ec='r')

    # vehicle axes
    plt.arrow(draw_vx.cg[i], draw_vy.cg[i], draw_vx.xaxis[i]-draw_vx.cg[i], draw_vy.xaxis[i]-draw_vy.cg[i], head_width=.5, head_length=0.5, fc='k', ec='k')
    plt.arrow(draw_vx.cg[i], draw_vy.cg[i], draw_vx.yaxis[i]-draw_vx.cg[i], draw_vy.yaxis[i]-draw_vy.cg[i], head_width=.5, head_length=0.5, fc='b', ec='b')
    plt.gca().invert_yaxis()
    plt.show()

def draw_vehicle_motion(draw_gx, draw_gy, draw_tire, i):
    #%% Plot Vehicle in Global reference frame

    bdy_x = (draw_gx.b_lfc[i], draw_gx.b_rfc[i], draw_gx.b_rrc[i], draw_gx.b_lrc[i], draw_gx.b_lfc[i])
    bdy_y = (draw_gy.b_lfc[i], draw_gy.b_rfc[i], draw_gy.b_rrc[i], draw_gy.b_lrc[i], draw_gy.b_lfc[i])

    lfw_x = (draw_gx.lfw_a[i], draw_gx.lfw_b[i], draw_gx.lfw_c[i], draw_gx.lfw_d[i], draw_gx.lfw_a[i])
    lfw_y = (draw_gy.lfw_a[i], draw_gy.lfw_b[i], draw_gy.lfw_c[i], draw_gy.lfw_d[i], draw_gy.lfw_a[i])

    rfw_x = (draw_gx.rfw_a[i], draw_gx.rfw_b[i], draw_gx.rfw_c[i], draw_gx.rfw_d[i], draw_gx.rfw_a[i])
    rfw_y = (draw_gy.rfw_a[i], draw_gy.rfw_b[i], draw_gy.rfw_c[i], draw_gy.rfw_d[i], draw_gy.rfw_a[i])

    rrw_x = (draw_gx.rrw_a[i], draw_gx.rrw_b[i], draw_gx.rrw_c[i], draw_gx.rrw_d[i], draw_gx.rrw_a[i])
    rrw_y = (draw_gy.rrw_a[i], draw_gy.rrw_b[i], draw_gy.rrw_c[i], draw_gy.rrw_d[i], draw_gy.rrw_a[i])

    lrw_x = (draw_gx.lrw_a[i], draw_gx.lrw_b[i], draw_gx.lrw_c[i], draw_gx.lrw_d[i], draw_gx.lrw_a[i])
    lrw_y = (draw_gy.lrw_a[i], draw_gy.lrw_b[i], draw_gy.lrw_c[i], draw_gy.lrw_d[i], draw_gy.lrw_a[i])


    fig = plt.figure(figsize=(15,10))
    #plt.xlim([-300, 50])
    #plt.ylim([-100, 100])

    plt.plot(bdy_x, bdy_y, 'k')
    plt.scatter(draw_gx.lfw[i], draw_gy.lfw[i], c='b')       # left front wheel center
    plt.plot(lfw_x, lfw_y, 'b')
    plt.scatter(draw_gx.rfw[i], draw_gy.rfw[i], c='g')       # right front wheel center
    plt.plot(rfw_x, rfw_y, 'g')
    plt.scatter(draw_gx.rrw[i], draw_gy.rrw[i], c='m')       # right rear wheel center
    plt.plot(rrw_x, rrw_y,'m')
    plt.scatter(draw_gx.lrw[i], draw_gy.lrw[i], c='orange')       # left rear wheel center
    plt.plot(lrw_x, lrw_y, 'orange')
    plt.scatter(draw_gx.cg[i], draw_gy.cg[i],s=100, c='k')                                # vehicle CG
    plt.arrow(draw_gx.cg[i], draw_gy.cg[i], draw_gx.vel_v[i]*0.2, draw_gy.vel_v[i]*0.2, head_width=.5, head_length=0.5, fc='r', ec='r')     # velocity vector
    plt.arrow(draw_gx.cg[i], draw_gy.cg[i], draw_gx.xaxis[i]-draw_gx.cg[i], draw_gy.xaxis[i]-draw_gy.cg[i], head_width=.5, head_length=0.5, fc='k', ec='k')     # vehicle axes
    plt.arrow(draw_gx.cg[i], draw_gy.cg[i], draw_gx.yaxis[i]-draw_gx.cg[i], draw_gy.yaxis[i]-draw_gy.cg[i], head_width=.5, head_length=0.5, fc='b', ec='b')     # vehicle axes

    # for loop up to i
    if draw_tire == 1:
        for i in range(0, i):
            if draw_gx.loc[i, 'lf_lock'] == 0:
                plt.scatter(draw_gx.loc[i, 'lfw'], draw_gy.loc[i, 'lfw'], c = 'b', s = 1, marker = '.')
            elif draw_gx.loc[i, 'lf_lock'] == 1:
                plt.scatter(draw_gx.loc[i, 'lfw'], draw_gy.loc[i, 'lfw'], c = 'b', s = 4, marker = 's')

            if draw_gx.loc[i, 'rf_lock'] == 0:
                plt.scatter(draw_gx.loc[i, 'rfw'], draw_gy.loc[i, 'rfw'], c = 'g', s = 1, marker = '.')
            elif draw_gx.loc[i, 'rf_lock'] == 1:
                plt.scatter(draw_gx.loc[i, 'rfw'], draw_gy.loc[i, 'rfw'], c = 'g', s = 4, marker = 's')

            if draw_gx.loc[i, 'rr_lock'] == 0:
                plt.scatter(draw_gx.loc[i, 'rrw'], draw_gy.loc[i, 'rrw'], c = 'm', s = 1, marker = '.')
            elif draw_gx.loc[i, 'rr_lock'] == 1:
                plt.scatter(draw_gx.loc[i, 'rrw'], draw_gy.loc[i, 'rrw'], c = 'm', s = 4, marker = 's')

            if draw_gx.loc[i, 'lr_lock'] == 0:
                plt.scatter(draw_gx.loc[i, 'lrw'], draw_gy.loc[i, 'lrw'], c = 'orange', s = 1, marker = '.')
            elif draw_gx.loc[i, 'lr_lock'] == 1:
                plt.scatter(draw_gx.loc[i, 'lrw'], draw_gy.loc[i, 'lrw'], c = 'orange', s = 4, marker = 's')

    plt.gca().invert_yaxis()
    plt.show()
    return fig

def draw_vehicle_motion_loop(draw_gx, draw_gy, draw_tire, ax, i):
    #%% Plot Vehicle in Global reference frame

    bdy_x = (draw_gx.b_lfc[i], draw_gx.b_rfc[i], draw_gx.b_rrc[i], draw_gx.b_lrc[i], draw_gx.b_lfc[i])
    bdy_y = (draw_gy.b_lfc[i], draw_gy.b_rfc[i], draw_gy.b_rrc[i], draw_gy.b_lrc[i], draw_gy.b_lfc[i])

    lfw_x = (draw_gx.lfw_a[i], draw_gx.lfw_b[i], draw_gx.lfw_c[i], draw_gx.lfw_d[i], draw_gx.lfw_a[i])
    lfw_y = (draw_gy.lfw_a[i], draw_gy.lfw_b[i], draw_gy.lfw_c[i], draw_gy.lfw_d[i], draw_gy.lfw_a[i])

    rfw_x = (draw_gx.rfw_a[i], draw_gx.rfw_b[i], draw_gx.rfw_c[i], draw_gx.rfw_d[i], draw_gx.rfw_a[i])
    rfw_y = (draw_gy.rfw_a[i], draw_gy.rfw_b[i], draw_gy.rfw_c[i], draw_gy.rfw_d[i], draw_gy.rfw_a[i])

    rrw_x = (draw_gx.rrw_a[i], draw_gx.rrw_b[i], draw_gx.rrw_c[i], draw_gx.rrw_d[i], draw_gx.rrw_a[i])
    rrw_y = (draw_gy.rrw_a[i], draw_gy.rrw_b[i], draw_gy.rrw_c[i], draw_gy.rrw_d[i], draw_gy.rrw_a[i])

    lrw_x = (draw_gx.lrw_a[i], draw_gx.lrw_b[i], draw_gx.lrw_c[i], draw_gx.lrw_d[i], draw_gx.lrw_a[i])
    lrw_y = (draw_gy.lrw_a[i], draw_gy.lrw_b[i], draw_gy.lrw_c[i], draw_gy.lrw_d[i], draw_gy.lrw_a[i])


  
    #plt.xlim([-300, 50])
    #plt.ylim([-100, 100])

    ax.plot(bdy_x, bdy_y, 'k')
    ax.scatter(draw_gx.lfw[i], draw_gy.lfw[i], c='b')       # left front wheel center
    ax.plot(lfw_x, lfw_y, 'b')
    ax.scatter(draw_gx.rfw[i], draw_gy.rfw[i], c='g')       # right front wheel center
    ax.plot(rfw_x, rfw_y, 'g')
    ax.scatter(draw_gx.rrw[i], draw_gy.rrw[i], c='m')       # right rear wheel center
    ax.plot(rrw_x, rrw_y,'m')
    ax.scatter(draw_gx.lrw[i], draw_gy.lrw[i], c='orange')       # left rear wheel center
    ax.plot(lrw_x, lrw_y, 'orange')
    ax.scatter(draw_gx.cg[i], draw_gy.cg[i],s=100, c='k')                                # vehicle CG
    ax.arrow(draw_gx.cg[i], draw_gy.cg[i], draw_gx.vel_v[i]*0.2, draw_gy.vel_v[i]*0.2, head_width=.5, head_length=0.5, fc='r', ec='r')     # velocity vector
    ax.arrow(draw_gx.cg[i], draw_gy.cg[i], draw_gx.xaxis[i]-draw_gx.cg[i], draw_gy.xaxis[i]-draw_gy.cg[i], head_width=.5, head_length=0.5, fc='k', ec='k')     # vehicle axes
    ax.arrow(draw_gx.cg[i], draw_gy.cg[i], draw_gx.yaxis[i]-draw_gx.cg[i], draw_gy.yaxis[i]-draw_gy.cg[i], head_width=.5, head_length=0.5, fc='b', ec='b')     # vehicle axes

    # for loop up to i
    if draw_tire == 1:
        for i in range(0, i):
            if draw_gx.loc[i, 'lf_lock'] == 0:
                ax.scatter(draw_gx.loc[i, 'lfw'], draw_gy.loc[i, 'lfw'], c = 'b', s = 1, marker = '.')
            elif draw_gx.loc[i, 'lf_lock'] == 1:
                ax.scatter(draw_gx.loc[i, 'lfw'], draw_gy.loc[i, 'lfw'], c = 'b', s = 4, marker = 's')

            if draw_gx.loc[i, 'rf_lock'] == 0:
                ax.scatter(draw_gx.loc[i, 'rfw'], draw_gy.loc[i, 'rfw'], c = 'g', s = 1, marker = '.')
            elif draw_gx.loc[i, 'rf_lock'] == 1:
                ax.scatter(draw_gx.loc[i, 'rfw'], draw_gy.loc[i, 'rfw'], c = 'g', s = 4, marker = 's')

            if draw_gx.loc[i, 'rr_lock'] == 0:
                ax.scatter(draw_gx.loc[i, 'rrw'], draw_gy.loc[i, 'rrw'], c = 'm', s = 1, marker = '.')
            elif draw_gx.loc[i, 'rr_lock'] == 1:
                ax.scatter(draw_gx.loc[i, 'rrw'], draw_gy.loc[i, 'rrw'], c = 'm', s = 4, marker = 's')

            if draw_gx.loc[i, 'lr_lock'] == 0:
                ax.scatter(draw_gx.loc[i, 'lrw'], draw_gy.loc[i, 'lrw'], c = 'orange', s = 1, marker = '.')
            elif draw_gx.loc[i, 'lr_lock'] == 1:
                ax.scatter(draw_gx.loc[i, 'lrw'], draw_gy.loc[i, 'lrw'], c = 'orange', s = 4, marker = 's')

    plt.gca().invert_yaxis()
    #plt.show()
    return

def draw_vehicle_motion_outline(draw_gx, draw_gy, draw_tire, i):
    #%% Plot Vehicle in Global reference frame

    bdy_x = (draw_gx.b_lfc[i], draw_gx.b_rfc[i], draw_gx.b_rrc[i], draw_gx.b_lrc[i], draw_gx.b_lfc[i])
    bdy_y = (draw_gy.b_lfc[i], draw_gy.b_rfc[i], draw_gy.b_rrc[i], draw_gy.b_lrc[i], draw_gy.b_lfc[i])

    fig = plt.figure(figsize=(15,10))
    plt.xlim([-10, 1000])
    plt.ylim([-300, 300])

    plt.plot(bdy_x, bdy_y, 'k')
    plt.scatter(draw_gx.lfw[i], draw_gy.lfw[i], c='b')       # left front wheel center
    plt.scatter(draw_gx.rfw[i], draw_gy.rfw[i], c='g')       # right front wheel center
    plt.scatter(draw_gx.rrw[i], draw_gy.rrw[i], c='m')       # right rear wheel center
    plt.scatter(draw_gx.lrw[i], draw_gy.lrw[i], c='orange')       # left rear wheel center
    plt.scatter(draw_gx.cg[i], draw_gy.cg[i],s=100, c='k')                                # vehicle CG
    plt.arrow(draw_gx.cg[i], draw_gy.cg[i], draw_gx.vel_v[i], draw_gy.vel_v[i], head_width=.5, head_length=0.5, fc='r', ec='r')     # velocity vector
    plt.arrow(draw_gx.cg[i], draw_gy.cg[i], draw_gx.xaxis[i]-draw_gx.cg[i], draw_gy.xaxis[i]-draw_gy.cg[i], head_width=.5, head_length=0.5, fc='k', ec='k')     # vehicle axes
    plt.arrow(draw_gx.cg[i], draw_gy.cg[i], draw_gx.yaxis[i]-draw_gx.cg[i], draw_gy.yaxis[i]-draw_gy.cg[i], head_width=.5, head_length=0.5, fc='b', ec='b')     # vehicle axes

    # for loop up to i
    if draw_tire == 1:
        for i in range(0, i):
            if draw_gx.loc[i, 'lf_lock'] == 0:
                plt.scatter(draw_gx.loc[i, 'lfw'], draw_gy.loc[i, 'lfw'], c = 'b', s = 1, marker = '.')
            elif draw_gx.loc[i, 'lf_lock'] == 1:
                plt.scatter(draw_gx.loc[i, 'lfw'], draw_gy.loc[i, 'lfw'], c = 'b', s = 4, marker = 's')

            if draw_gx.loc[i, 'rf_lock'] == 0:
                plt.scatter(draw_gx.loc[i, 'rfw'], draw_gy.loc[i, 'rfw'], c = 'g', s = 1, marker = '.')
            elif draw_gx.loc[i, 'rf_lock'] == 1:
                plt.scatter(draw_gx.loc[i, 'rfw'], draw_gy.loc[i, 'rfw'], c = 'g', s = 4, marker = 's')

            if draw_gx.loc[i, 'rr_lock'] == 0:
                plt.scatter(draw_gx.loc[i, 'rrw'], draw_gy.loc[i, 'rrw'], c = 'm', s = 1, marker = '.')
            elif draw_gx.loc[i, 'rr_lock'] == 1:
                plt.scatter(draw_gx.loc[i, 'rrw'], draw_gy.loc[i, 'rrw'], c = 'm', s = 4, marker = 's')

            if draw_gx.loc[i, 'lr_lock'] == 0:
                plt.scatter(draw_gx.loc[i, 'lrw'], draw_gy.loc[i, 'lrw'], c = 'orange', s = 1, marker = '.')
            elif draw_gx.loc[i, 'lr_lock'] == 1:
                plt.scatter(draw_gx.loc[i, 'lrw'], draw_gy.loc[i, 'lrw'], c = 'orange', s = 4, marker = 's')

    plt.gca().invert_yaxis()
    plt.show()
    return fig


def draw_vehicles_impact(draw_gx1, draw_gy1, draw_gx2, draw_gy2, thetac):
    draw_gx = draw_gx1.append(draw_gx2, ignore_index = True)
    draw_gy = draw_gy1.append(draw_gy2, ignore_index = True)

    dx_t = 20 * math.cos(thetac)
    dy_t = 20 * math.sin(thetac)
    dx_n = 10 * math.cos(thetac+90*math.pi/180)
    dy_n = 10 * math.sin(thetac+90*math.pi/180)

    fig = plt.figure(figsize=(15,10))

    for i in (range(len(draw_gx))):
        #%% Plot Vehicle in Global reference frame
        bdy_x = (draw_gx.b_lfc[i], draw_gx.b_rfc[i], draw_gx.b_rrc[i], draw_gx.b_lrc[i], draw_gx.b_lfc[i])
        bdy_y = (draw_gy.b_lfc[i], draw_gy.b_rfc[i], draw_gy.b_rrc[i], draw_gy.b_lrc[i], draw_gy.b_lfc[i])

        lfw_x = (draw_gx.lfw_a[i], draw_gx.lfw_b[i], draw_gx.lfw_c[i], draw_gx.lfw_d[i], draw_gx.lfw_a[i])
        lfw_y = (draw_gy.lfw_a[i], draw_gy.lfw_b[i], draw_gy.lfw_c[i], draw_gy.lfw_d[i], draw_gy.lfw_a[i])

        rfw_x = (draw_gx.rfw_a[i], draw_gx.rfw_b[i], draw_gx.rfw_c[i], draw_gx.rfw_d[i], draw_gx.rfw_a[i])
        rfw_y = (draw_gy.rfw_a[i], draw_gy.rfw_b[i], draw_gy.rfw_c[i], draw_gy.rfw_d[i], draw_gy.rfw_a[i])

        rrw_x = (draw_gx.rrw_a[i], draw_gx.rrw_b[i], draw_gx.rrw_c[i], draw_gx.rrw_d[i], draw_gx.rrw_a[i])
        rrw_y = (draw_gy.rrw_a[i], draw_gy.rrw_b[i], draw_gy.rrw_c[i], draw_gy.rrw_d[i], draw_gy.rrw_a[i])

        lrw_x = (draw_gx.lrw_a[i], draw_gx.lrw_b[i], draw_gx.lrw_c[i], draw_gx.lrw_d[i], draw_gx.lrw_a[i])
        lrw_y = (draw_gy.lrw_a[i], draw_gy.lrw_b[i], draw_gy.lrw_c[i], draw_gy.lrw_d[i], draw_gy.lrw_a[i])


        plt.plot(bdy_x, bdy_y, 'k')
        plt.scatter(draw_gx.lfw[i], draw_gy.lfw[i], c='b')       # left front wheel center
        plt.plot(lfw_x, lfw_y, 'b')
        plt.scatter(draw_gx.rfw[i], draw_gy.rfw[i], c='g')       # right front wheel center
        plt.plot(rfw_x, rfw_y, 'g')
        plt.scatter(draw_gx.rrw[i], draw_gy.rrw[i], c='m')       # right rear wheel center
        plt.plot(rrw_x, rrw_y,'m')
        plt.scatter(draw_gx.lrw[i], draw_gy.lrw[i], c='orange')       # left rear wheel center
        plt.plot(lrw_x, lrw_y, 'orange')

        if i == 0:
            plt.scatter(draw_gx.cg[i], draw_gy.cg[i],s=100, c='b')                                  # vehicle 1 CG
        if i == 1:
            plt.scatter(draw_gx.cg[i], draw_gy.cg[i],s=100, c='g')                                  # vehicle 2 CG

        plt.scatter(draw_gx.poi[i], draw_gy.poi[i],s=100, c='r')                                # POI
        plt.arrow(draw_gx.cg[i], draw_gy.cg[i], draw_gx.vel_v[i]-draw_gx.cg[i], draw_gy.vel_v[i]-draw_gy.cg[i], head_width=.5, head_length=0.5, fc='r', ec='r')     # velocity vector
        plt.arrow(draw_gx.cg[i], draw_gy.cg[i], draw_gx.xaxis[i]-draw_gx.cg[i], draw_gy.xaxis[i]-draw_gy.cg[i], head_width=.5, head_length=0.5, fc='k', ec='k')     # vehicle axes
        plt.arrow(draw_gx.cg[i], draw_gy.cg[i], draw_gx.yaxis[i]-draw_gx.cg[i], draw_gy.yaxis[i]-draw_gy.cg[i], head_width=.5, head_length=0.5, fc='b', ec='b')     # vehicle axes

    plt.arrow(draw_gx.poi[i], draw_gy.poi[i], dx_t, dy_t, head_width=.8, head_length=0.8, fc='k', ec='r', linestyle = '--')     # collision plane - tangent
    plt.arrow(draw_gx.poi[i], draw_gy.poi[i], dx_n, dy_n, head_width=.8, head_length=0.8, fc='r', ec='r', linestyle = '--')     # collision plane - normal
    plt.xlim([-50, 50])
    plt.ylim([-30, 30])
    plt.gca().invert_yaxis()
    plt.show()
    return fig

def draw_impact_motion(draw_gx1, draw_gy1, draw_gx2, draw_gy2, draw_tire, i):

    #%% Plot Vehicle in Global reference frame
    fig = plt.figure(figsize=(15,10))

    for j in range(0,2):
        if j == 0:
            draw_gx = draw_gx1
            draw_gy = draw_gy1
        elif j == 1:
            draw_gx = draw_gx2
            draw_gy = draw_gy2

        for i in [0,i]:
            bdy_x = (draw_gx.b_lfc[i], draw_gx.b_rfc[i], draw_gx.b_rrc[i], draw_gx.b_lrc[i], draw_gx.b_lfc[i])
            bdy_y = (draw_gy.b_lfc[i], draw_gy.b_rfc[i], draw_gy.b_rrc[i], draw_gy.b_lrc[i], draw_gy.b_lfc[i])

            lfw_x = (draw_gx.lfw_a[i], draw_gx.lfw_b[i], draw_gx.lfw_c[i], draw_gx.lfw_d[i], draw_gx.lfw_a[i])
            lfw_y = (draw_gy.lfw_a[i], draw_gy.lfw_b[i], draw_gy.lfw_c[i], draw_gy.lfw_d[i], draw_gy.lfw_a[i])

            rfw_x = (draw_gx.rfw_a[i], draw_gx.rfw_b[i], draw_gx.rfw_c[i], draw_gx.rfw_d[i], draw_gx.rfw_a[i])
            rfw_y = (draw_gy.rfw_a[i], draw_gy.rfw_b[i], draw_gy.rfw_c[i], draw_gy.rfw_d[i], draw_gy.rfw_a[i])

            rrw_x = (draw_gx.rrw_a[i], draw_gx.rrw_b[i], draw_gx.rrw_c[i], draw_gx.rrw_d[i], draw_gx.rrw_a[i])
            rrw_y = (draw_gy.rrw_a[i], draw_gy.rrw_b[i], draw_gy.rrw_c[i], draw_gy.rrw_d[i], draw_gy.rrw_a[i])

            lrw_x = (draw_gx.lrw_a[i], draw_gx.lrw_b[i], draw_gx.lrw_c[i], draw_gx.lrw_d[i], draw_gx.lrw_a[i])
            lrw_y = (draw_gy.lrw_a[i], draw_gy.lrw_b[i], draw_gy.lrw_c[i], draw_gy.lrw_d[i], draw_gy.lrw_a[i])

            plt.plot(bdy_x, bdy_y, 'k')
            plt.scatter(draw_gx.lfw[i], draw_gy.lfw[i], c='b')       # left front wheel center
            plt.plot(lfw_x, lfw_y, 'b')
            plt.scatter(draw_gx.rfw[i], draw_gy.rfw[i], c='g')       # right front wheel center
            plt.plot(rfw_x, rfw_y, 'g')
            plt.scatter(draw_gx.rrw[i], draw_gy.rrw[i], c='m')       # right rear wheel center
            plt.plot(rrw_x, rrw_y,'m')
            plt.scatter(draw_gx.lrw[i], draw_gy.lrw[i], c='orange')       # left rear wheel center
            plt.plot(lrw_x, lrw_y, 'orange')
            plt.scatter(draw_gx.cg[i], draw_gy.cg[i],s=100, c='k')                                # vehicle CG
            plt.arrow(draw_gx.cg[i], draw_gy.cg[i], draw_gx.vel_v[i]-draw_gx.cg[i], draw_gy.vel_v[i]-draw_gy.cg[i], head_width=.5, head_length=0.5, fc='r', ec='r')     # velocity vector
            plt.arrow(draw_gx.cg[i], draw_gy.cg[i], draw_gx.xaxis[i]-draw_gx.cg[i], draw_gy.xaxis[i]-draw_gy.cg[i], head_width=.5, head_length=0.5, fc='k', ec='k')     # vehicle axes
            plt.arrow(draw_gx.cg[i], draw_gy.cg[i], draw_gx.yaxis[i]-draw_gx.cg[i], draw_gy.yaxis[i]-draw_gy.cg[i], head_width=.5, head_length=0.5, fc='b', ec='b')     # vehicle axes

        # for loop up to i
        if draw_tire == 1:
            for i in range(0, i):
                if draw_gx.loc[i, 'lf_lock'] == 0:
                    plt.scatter(draw_gx.loc[i, 'lfw'], draw_gy.loc[i, 'lfw'], c = 'b', s = 1, marker = '.')
                elif draw_gx.loc[i, 'lf_lock'] == 1:
                    plt.scatter(draw_gx.loc[i, 'lfw'], draw_gy.loc[i, 'lfw'], c = 'b', s = 4, marker = 's')

                if draw_gx.loc[i, 'rf_lock'] == 0:
                    plt.scatter(draw_gx.loc[i, 'rfw'], draw_gy.loc[i, 'rfw'], c = 'g', s = 1, marker = '.')
                elif draw_gx.loc[i, 'rf_lock'] == 1:
                    plt.scatter(draw_gx.loc[i, 'rfw'], draw_gy.loc[i, 'rfw'], c = 'g', s = 4, marker = 's')

                if draw_gx.loc[i, 'rr_lock'] == 0:
                    plt.scatter(draw_gx.loc[i, 'rrw'], draw_gy.loc[i, 'rrw'], c = 'm', s = 1, marker = '.')
                elif draw_gx.loc[i, 'rr_lock'] == 1:
                    plt.scatter(draw_gx.loc[i, 'rrw'], draw_gy.loc[i, 'rrw'], c = 'm', s = 4, marker = 's')

                if draw_gx.loc[i, 'lr_lock'] == 0:
                    plt.scatter(draw_gx.loc[i, 'lrw'], draw_gy.loc[i, 'lrw'], c = 'orange', s = 1, marker = '.')
                elif draw_gx.loc[i, 'lr_lock'] == 1:
                    plt.scatter(draw_gx.loc[i, 'lrw'], draw_gy.loc[i, 'lrw'], c = 'orange', s = 4, marker = 's')


    plt.xlim([-300, 50])
    plt.ylim([-100, 100])
    plt.gca().invert_yaxis()
    plt.show()
    return fig



"""
repository for all variables used in pycrash modulues
"""
definitions = {
'delta_deg':'tire steer angle (degrees)',
'delta_rad':'tire steer angle (radians)',
'beta_deg':'velocity vectory in intertial reference frame (degrees)',
'beta_rad':'velocity vectory in intertial reference frame (radians)',
'lf_alpha':'left front tire slip angle in tire reference frame (radians)',

}



# TODO: check properties of each input - use list of data types
# TODO: option to save inputs and outputs to csv

# %% modules
from .visualization.vehicle import plot_driver_inputs
import pandas as pd
import numpy as np
import os
import csv

project_dir = os.path.dirname(os.getcwd())
input_dir = os.path.join(project_dir, 'data', 'input')

# load defaults
sim_defaults = {'dt_motion': 0.01,
                'mu_max': 0.76,
                'alpha_max': 0.174533}

mu_max = sim_defaults['mu_max']    # maximum available friction
dt_motion = sim_defaults['dt_motion']            # iteration time step

print('Current values for defined constants:')
print(f'maximum available friction (mu_max) = {mu_max}')
print(f'time step for vehicle motion (dt) = {dt_motion} s')

# vehicle inputs - values used in csv file for input should match these below
# additional values requires a value for input_query, veh_input, dtype be provided
# must be in same position in each list

input_query = ["Model year",
"Vehicle make",
"Vehicle model",
"Vehicle weight (lb)",
"Vehicle Identification Number (VIN)",
"Percent Braking",
"Steering ratio",
"Initial X position (ft)",
"Initial Y position (ft)",
"Initial heading angle (deg)",
"Vehicle width (ft)",
"Vehicle length (ft)",
"CG height (ft)",
"CG to front axle (ft)",
"CG to rear axle (ft)",
"Wheelbase (ft)",
"Track width (ft)",
"Front overhang (ft)",
"Rear overhang (ft)",
"Tire diameter (ft)",
"Tire width (ft)",
"Izz (lb-ft-s^2)",
"Front wheel drive (0/1)",
"Rear wheel drive (0/1)",
"All wheel drive (0/1)",
"Stiffness value A [lb/in]",
"Stiffness slope B [lb/in/in]",
"Effective spring stiffness k [lb/in]",
"Damage length L [in]",
"Crush depth c [in]",
"Initial forward velocity Vx (mph)",
"Initial lateral velocity Vy (mph)",
"Initial Yaw Rate (deg/s)"]

veh_inputs = ["year",
"make",
"model",
"weight",
"vin",
"brake",
"steer_ratio",
"init_x_pos",
"init_y_pos",
"head_angle",
"width",
"length",
"hcg",
"lcgf",
"lcgr",
"wb",
"track",
"f_hang",
"r_hang",
"tire_d",
"tire_w",
"izz",
"fwd",
"rwd",
"awd",
"A",
"B",
"k",
"L",
"c",
"vx_initial",
"vy_initial",
"omega_z",
"striking"]




class Vehicle:
    """
    Vehicle - contains all data assigned to a vehicle used to run various simulations
    not all values are required to create a vehicle instance
    requires 'Name' - used to idenify vehicle in simulations
    can be useful to create mutiple iterations of the same vehicles
    Veh1_Weight1, Veh1_Weight2, etc.
    """

    # TODO: create a loop through the input dictionary and assign to class

    def __init__(self, name, input_dict=None):
        self.name = str(name)
        self.type = "vehicle"   # class type for reference

        if input_dict != None:
            for key, value in input_dict.items():
                if key in veh_inputs:
                    if key in ['make', 'model', 'vin']:
                        setattr(self, key, str(value))
                    else:
                        setattr(self, key, float(value))
                else:
                    print(f"Input entry {key} unknown, setting to {value}")
                    setattr(self, key, float(value))

            print(f'Vehicle inputs for {self.name} applied succesfully')

    def manual_specs(self):  # loop through lists above to create inputs
        for i in range(len(input_query)):
                userEntry = input(input_query[i])
                try:
                    setattr(self, veh_inputs[i], float(userEntry))  # convert to float if possible
                except:
                    setattr(self, veh_inputs[i], userEntry)
                print(f'{input_query[i]} = {userEntry}')

    def load_specs(self, filename):
        """ provide file name to .csv file with defined layout
            file must be located in "input" directory
            uses contents of csv to determine attributes for vehicle variables
            altering the key names will break functionality of the simulation
        """
        with open(os.path.join(os.getcwd(), "data", "input", filename)) as csvfile:
            readCSV = csv.reader(csvfile, delimiter=',')
            for row in readCSV:
                print(f'{row[0]} = {row[2]}')
                try:
                    setattr(self, row[1], float(row[2]))  # convert to float if possible
                except:
                    setattr(self, row[1], row[2])

    def input_dict(self):
        vehicle_input_dict = {"year":self.year,
        "make":self.make,
        "model":self.model,
        "weight":self.weight,
        "vin":self.vin,
        "brake":self.brake,
        "steer_ratio":self.steer_ratio,
        "init_x_pos":self ,
        "init_y_pos":self.init_y_pos,
        "head_angle":self.head_angle,
        "width":self.width,
        "length":self.length,
        "hcg":self.hcg,
        "lcgf":self.lcgf,
        "lcgr":self.lcgr,
        "wb":self.wb,
        "track":self.track,
        "f_hang":self.f_hang,
        "r_hang":self.r_hang,
        "tire_d":self.tire_d,
        "tire_w":self.tire_w,
        "izz":self.izz,
        "fwd":self.fwd,
        "rwd":self.rwd,
        "awd":self.awd,
        "A":self.A,
        "B":self.B,
        "k":self.k,
        "L":self.L,
        "c":self.c,
        "vx_initial":self.vx_initial,
        "vy_initial":self.vy_initial,
        "omega_z":self.omega_z}

        return vehicle_input_dict

    def time_inputs(self, time, throttle, brake, steer, show_plot=True):
        """
        Driver inputs | time (s) | throttle (%) | braking (%) | steering (deg) |
        time step can be user defined, inputs will be interpolated to match dt for simulation
        user create data frame with the necessary columns
        """
        if len(time) == 0:
            print('Input data for time has zero length')
            print('No driver input applied to vehicle')
        else:
            inputdf = pd.DataFrame(list(zip(throttle, brake, steer)), columns = ['throttle', 'brake', 'steer'])
            t = list(np.arange(0, max(time) + dt_motion, dt_motion)) # create time array from 0 to max time in inputs, does not mean simulation will stop at this time_inputs
            t = [float(i) for i in t]
            df = pd.DataFrame()                                                           # create dataframe for vehicle input with interpolated values
            df['t'] = t
            inputdf['input_t'] = [float(num) for num in time]
            df.t = df.t.round(3).astype(float)
            df = pd.merge(df, inputdf, how = 'left', left_on = 't', right_on = 'input_t')  # merge input data with time data at specified time step
            df = df.interpolate(method = 'linear', axis = 0) # interpolate NaN values left after merging
            df.drop(columns = ['input_t', 't'], inplace = True)  # drop input time column
            df['t'] = t # reset time column due to interpolating
            df['t'] = df.t.round(3) # reset signficant digits
            df = df.reset_index(drop = True)
            self.driver_input = df
            print(f'Driver inputs applied to {self.name}')
            if show_plot:
                plot_driver_inputs(self)

    def read_time_inputsCSV(self, filename):
        """
        Driver inputs | time (s) | throttle (%) | brake (%) | steer (deg) |
        time step can be user defined, inputs will be interpolated to match dt for simulation
        reads data from csv file
        will override other inputs applied to vehicle
        filename should include .csv - "example_file_name.csv"
        """
        header_list = ["time", "throttle", "brake", "steer"]
        if os.path.isfile(os.path.join(input_dir, filename)):
            time_inputs = pd.read_csv(os.path.join(input_dir, filename), skiprows=1, header=None, names = header_list)
            time_inputs = time_inputs.astype(float)
            if len(time_inputs) == 0:
                print('Time input file appears blank')
            else:
                t = list(np.arange(0, dt_motion+time_inputs.loc[len(time_inputs.time)-1, 'time'], dt_motion))  # create time array from 0 to max time in inputs, this will be end time for simulation
                df = pd.DataFrame()                                                           # create dataframe for vehicle input with interpolated values
                df['t'] = t
                time_inputs['input_t'] = time_inputs.time.round(3)
                df.t = df.t.round(3)
                df = pd.merge(df, time_inputs, how = 'left', left_on = 't', right_on = 'input_t') # merge input data with time data at specified time step
                df = df.interpolate(method = 'linear') # interpolate NaN values left after merging
                df.drop(columns = ['input_t', 't'], inplace = True)  # drop input time column
                df['t'] = t # reset time column due to interpolating
                df['t'] = df.t.round(3) # reset signficant digits
                df = df.reset_index(drop = True)
                self.driver_input = df
                print(f'Driver inputs applied to {self.name}')
        else:
            print('No Time Input File Provided')

        plot_driver_inputs(self)

    def plot_driver_inputs(self):
            plot_driver_inputs(self)

    # TODO: create ability to change inputs by distance
    def dist_inputs(self, filename):
        """
        Driver inputs | vehicle travel distance (ft) | brake (%) | steer (deg) |
        will override other inputs applied to vehicle
        """

    def show(self):
        for key in self.__dict__.keys():
            if isinstance(self.__dict__[key], pd.DataFrame):
                print("")
                print(f"First five rows of {key}:")
                print(self.__dict__[key].head())
            else:
                print(f"{key} -> {self.__dict__[key]}")



import plotly.graph_objects as go
import plotly.io as pio

pio.renderers.default = "plotly_mimetype"  # <- determines how plots are displayed using Plotly
# "browser"  # <- determines how plots are displayed using Plotly


def plot_TotalFdx(df):
    """ dataframe with at least two columns
    Disp - vehicle displacment in feet
    TotalForce - barrier force in lb
    """
    fig = go.Figure()
    fig.add_trace(go.Scatter(x=df.Disp,
                             y=df.TotalForce * -1,
                             mode='lines',
                             name="Total Force",
                             line=dict(color="rgb(52, 64, 235)", width=2)
                             ))
    fig.update_layout(
        autosize=False,
        width=1000,
        height=550,
        title=f'Load Cell Barrier Force-Displacement Data',
        template='plotly_white',
        xaxis=dict(showgrid=False, title='Displacement (feet)'),
        font=dict(family='Arial', size=24, color='black'))

    fig.update_layout(showlegend=False)
    fig.update_yaxes(showgrid=False, showline=True, linewidth=1, linecolor='black', ticks="outside",
                     tickwidth=1, tickcolor='black', ticklen=10, zeroline=False, title_text='Force (lb)')
    fig.update_xaxes(showline=True, linewidth=1, linecolor='black', ticks="outside",
                     tickwidth=1, tickcolor='black', ticklen=10, zeroline=False)
    fig.show()



"""
plotting functions for vehicle model data
"""
import plotly.graph_objects as go
import numpy as np
from plotly.subplots import make_subplots
import plotly.io as pio
pio.renderers.default = "browser"

# TODO: create input for figure size - loads from "defaults" environmental variable?
width = 900
aspect_ratio = 16 / 9
figure_size = (width, width / aspect_ratio)
wheel_colors = ['rgb(0, 0, 255)', 'rgb(0, 255, 0)', 'rgb(153, 0, 204)', 'rgb(255, 102, 0)']
font_size = 24
tick_font_size = 22

def get_impactNum(i, impactIndex):
    if i < max(impactIndex):
        for key, value in impactIndex.items():
            if i <= key:
                return(value)
    else:
        return max(impactIndex.values())

def plot_impact(veh_list, i, impactIndex, tire_path=True, show_vector=False):
    fig = go.Figure()
    plotImpactPoint = get_impactNum(i, impactIndex)
    for veh in veh_list:
        # Body and tire outlines
        bdy_x = (veh.p_gx.b_lfc[i], veh.p_gx.b_rfc[i], veh.p_gx.b_rrc[i], veh.p_gx.b_lrc[i], veh.p_gx.b_lfc[i])
        bdy_y = (veh.p_gy.b_lfc[i], veh.p_gy.b_rfc[i], veh.p_gy.b_rrc[i], veh.p_gy.b_lrc[i], veh.p_gy.b_lfc[i])

        lfw_x = (veh.p_gx.lfw_a[i], veh.p_gx.lfw_b[i], veh.p_gx.lfw_c[i], veh.p_gx.lfw_d[i], veh.p_gx.lfw_a[i])
        lfw_y = (veh.p_gy.lfw_a[i], veh.p_gy.lfw_b[i], veh.p_gy.lfw_c[i], veh.p_gy.lfw_d[i], veh.p_gy.lfw_a[i])

        rfw_x = (veh.p_gx.rfw_a[i], veh.p_gx.rfw_b[i], veh.p_gx.rfw_c[i], veh.p_gx.rfw_d[i], veh.p_gx.rfw_a[i])
        rfw_y = (veh.p_gy.rfw_a[i], veh.p_gy.rfw_b[i], veh.p_gy.rfw_c[i], veh.p_gy.rfw_d[i], veh.p_gy.rfw_a[i])

        rrw_x = (veh.p_gx.rrw_a[i], veh.p_gx.rrw_b[i], veh.p_gx.rrw_c[i], veh.p_gx.rrw_d[i], veh.p_gx.rrw_a[i])
        rrw_y = (veh.p_gy.rrw_a[i], veh.p_gy.rrw_b[i], veh.p_gy.rrw_c[i], veh.p_gy.rrw_d[i], veh.p_gy.rrw_a[i])

        lrw_x = (veh.p_gx.lrw_a[i], veh.p_gx.lrw_b[i], veh.p_gx.lrw_c[i], veh.p_gx.lrw_d[i], veh.p_gx.lrw_a[i])
        lrw_y = (veh.p_gy.lrw_a[i], veh.p_gy.lrw_b[i], veh.p_gy.lrw_c[i], veh.p_gy.lrw_d[i], veh.p_gy.lrw_a[i])

        # body outline
        fig.add_trace(go.Scatter(x = bdy_x, y = bdy_y,
                            mode = 'lines',
                            name = 'body',
                            line = dict(color = 'rgb(0, 0, 0)', width = 2)))
        # wheel outline
        fig.add_trace(go.Scatter(x = lfw_x, y = lfw_y,
                            mode = 'lines',
                            name = 'LF',
                            line = dict(color = 'rgb(0, 0, 255)', width = 1)))
        fig.add_trace(go.Scatter(x = rfw_x, y = rfw_y,
                            mode = 'lines',
                            name = 'RF',
                            line = dict(color = 'rgb(0, 255, 0)', width = 1)))
        fig.add_trace(go.Scatter(x = rrw_x, y = rrw_y,
                            mode = 'lines',
                            name = 'RR',
                            line = dict(color = 'rgb(153, 0, 204)', width = 1)))
        fig.add_trace(go.Scatter(x = lrw_x, y = lrw_y,
                            mode = 'lines',
                            name = 'LR',
                            line = dict(color = 'rgb(255, 102, 0)', width = 1)))
        # CG
        cgx = [veh.p_gx.cg[i], veh.p_gx.cg[i]]
        cgy = [veh.p_gy.cg[i], veh.p_gy.cg[i]]
        time = [veh.model.t[i], veh.model.t[i]]

        fig.add_trace(go.Scatter(x = cgx, y = cgy,
                            mode = 'markers',
                            name = 'CG',
                            marker = dict(color = 'rgb(0, 0, 0)', size = 10),
                            customdata = time,
                            hovertemplate = '<b>x</b>:%{x:.1f}<br>' + '<b>Y</b>: %{y:.1f}<br>' + '<b>time</b>: %{customdata:.3f}'))

        # velocity vector
        if show_vector:
            fig.add_annotation(x = veh.p_gx.vel_v[i],
                               y = veh.p_gy.vel_v[i],
                               ax = veh.p_gx.cg[i],
                               ay = veh.p_gy.cg[i],
                               axref = 'x',
                               ayref = 'y',
                               xref = 'x',
                               yref = 'y',
                               text = "",
                               showarrow = True,
                               arrowsize = 2.5,
                               arrowhead = 1,
                               arrowwidth = 1.7,
                               arrowcolor = 'rgb(255, 0, 0)')

        # x axis
        fig.add_annotation(x = veh.p_gx.xaxis[i],
                           y = veh.p_gy.xaxis[i],
                           ax = veh.p_gx.cg[i],
                           ay = veh.p_gy.cg[i],
                           axref = 'x',
                           ayref = 'y',
                           xref = 'x',
                           yref = 'y',
                           text = "",
                           showarrow = True,
                           arrowsize = 2,
                           arrowhead = 1,
                           arrowwidth = 1.5,
                           arrowcolor = 'rgb(0, 0, 0)')

        # y axis
        fig.add_annotation(x = veh.p_gx.yaxis[i],
                           y = veh.p_gy.yaxis[i],
                           ax = veh.p_gx.cg[i],
                           ay = veh.p_gy.cg[i],
                           axref = 'x',
                           ayref = 'y',
                           xref = 'x',
                           yref = 'y',
                           text = "",
                           showarrow = True,
                           arrowsize = 2,
                           arrowhead = 1,
                           arrowwidth = 1.5,
                           arrowcolor = 'rgb(0, 0, 255)')

        # tire center plot depends on lock condition
        def setmarker(x):
            if x == 0:
                return 'circle'
            elif x == 1:
                return 'square-x'

        # plotly requires at least two rows to plot, so if i == 0, it will be set to 1
        gx = veh.p_gx[0:i]
        gy = veh.p_gy[0:i]
        #print(gx)

        if (tire_path):
            fig.add_trace(go.Scatter(x = gx.lfw, y = gy.lfw,
                                mode = 'markers',
                                name = 'LF',
                                marker = dict(color = 'rgb(0, 0, 255)', size = 2,
                                symbol = list(map(setmarker, gx.lf_lock)))))

            fig.add_trace(go.Scatter(x = gx.rfw, y = gy.rfw,
                                mode = 'markers',
                                name = 'RF',
                                marker = dict(color = 'rgb(0, 255, 0)', size = 2,
                                symbol = list(map(setmarker, gx.rf_lock)))))

            fig.add_trace(go.Scatter(x = gx.rrw, y = gy.rrw,
                                mode = 'markers',
                                name = 'RR',
                                marker = dict(color = 'rgb(153, 0, 204)', size = 2,
                                symbol = list(map(setmarker, gx.rr_lock)))))

            fig.add_trace(go.Scatter(x = gx.lrw, y = gy.lrw,
                                mode = 'markers',
                                name = 'LR',
                                marker = dict(color = 'rgb(255, 102, 0)', size = 2,
                                symbol = list(map(setmarker, gx.lr_lock)))))

        if veh.striking:
            # plot each impact plane
            fig.add_annotation(x=veh.p_gx[f'impact_{plotImpactPoint}_norm'][i],
                               y=veh.p_gy[f'impact_{plotImpactPoint}_norm'][i],
                               axref='x',
                               ayref='y',
                               text="",
                               showarrow=True,
                               ax=veh.p_gx[f'POI_{plotImpactPoint}'][i],
                               ay=veh.p_gy[f'POI_{plotImpactPoint}'][i],
                               arrowsize=2,
                               arrowhead=2,
                               arrowwidth=1.5,
                               arrowcolor='rgb(153, 0, 51)')
            # tangential impact plane
            fig.add_annotation(x=veh.p_gx[f'impact_{plotImpactPoint}_tang'][i],
                               y=veh.p_gy[f'impact_{plotImpactPoint}_tang'][i],
                               axref='x',
                               ayref='y',
                               text="",
                               showarrow=True,
                               ax=veh.p_gx[f'POI_{plotImpactPoint}'][i],
                               ay=veh.p_gy[f'POI_{plotImpactPoint}'][i],
                               arrowsize=1,
                               arrowhead=1,
                               arrowwidth=1,
                               arrowcolor='rgb(153, 0, 51)')

            fig.add_trace(go.Scatter(x=[veh.p_gx[f'POI_{plotImpactPoint}'][i], veh.p_gx[f'POI_{plotImpactPoint}'][i]],
                                     y=[veh.p_gy[f'POI_{plotImpactPoint}'][i], veh.p_gy[f'POI_{plotImpactPoint}'][i]],
                                     mode='markers',
                                     name=f'POI: {plotImpactPoint}',
                                     marker=dict(color='rgb(153, 0, 51)', size=7),
                                     ))

    # adjust axes to keep aspect aspect ratio
    """
    dx_max = veh.p_gx.cg.max() + 15
    dx_min = veh.p_gx.cg.min() - 15
    dy_max = veh.p_gy.cg.max() + 15
    dy_min = veh.p_gy.cg.min() - 15
    """
    dx_max = 15
    dx_min = -15
    dy_max = 10
    dy_min = -35

    dx = dx_max - dx_min
    dy = dy_max - dy_min

    if dx > dy:
        adj_x = aspect_ratio * dy / dx
        adj_y = 1
        #print(f" dx > dy -> adj_x = {adj_x}, adj_y = {adj_y}")
        dx_min = round(dx_min * adj_x)
        dx_max = round(dx_max * adj_x)
        #print(f"dx_min = {dx_min}, dx_max = {dx_max}")
        #print(f"dy_min = {dy_min}, dy_max = {dy_max}")
    else:
        adj_x = 1
        adj_y = (1 / aspect_ratio) * dx / dy
        #print(f" dy > dx -> adj_x = {adj_x}, adj_y = {adj_y}")
        dy_min = round(dy_min * adj_y)
        dy_max = round(dy_max * adj_y)
        #print(f"dx_min = {dx_min}, dx_max = {dx_max}")
        #print(f"dy_min = {dy_min}, dy_max = {dy_max}")

    fig.update_layout(
        showlegend = False,
        autosize = False,
        width = width,
        height = width / aspect_ratio,
        title = 'Vehicle Motion in Global Reference Frame',
        template = 'plotly_white',
        xaxis = dict(showgrid = False, title = 'x-axis - Forward (ft)', range = [2*dx_min, 2*dx_max], constrain="domain"),
        yaxis = dict(showgrid = False, title = 'y-axis - Rightward (ft)', scaleanchor = "x", scaleratio = 1),
        font = dict(family = 'Arial', size = font_size, color = 'black'))

    fig.update_xaxes(showline=True, linewidth=1, linecolor='black', ticks="outside",
                     tickwidth=1, tickcolor='black', ticklen=10, zeroline=False, tickfont=dict(size=tick_font_size))
    fig.update_yaxes(autorange="reversed", showline=True, linewidth=1, linecolor='black', ticks="outside",
                     tickwidth=1, tickcolor='black', ticklen=10, zeroline=False, tickfont=dict(size=tick_font_size))

    fig.show()



import pandas as pd
import math
import os

def impc(i, impactNum, poi_veh2x, poi_veh2y, impPointEdge, vehicle_list,
         strikingVehicle, struckVehicle, impc_inputs, dt_motion, show_results=True):
    cor = impc_inputs['cor']
    cof = impc_inputs['vehicle_mu']
    veh1 = vehicle_list[strikingVehicle]
    veh2 = vehicle_list[struckVehicle]
    print('<--- IMPC Model --->')
    print(f"Striking vehicle: {veh1.name}")
    print(f"Struck vehicle: {veh2.name}")

   # heading angle of tangent impact direction in global frame
    theta_c = veh1.model.theta_rad[i] + (impPointEdge[impactNum]['impact_points'][2] * math.pi / 180) + (90 / 180 * math.pi)

    # carpenter + welcher model
    # get cosine / sine results for coordinate transformation to the normal - tangent axis
    c1 = math.cos(veh1.model.theta_rad[i] - theta_c)
    s1 = math.sin(veh1.model.theta_rad[i] - theta_c)
    c2 = math.cos(veh2.model.theta_rad[i] - theta_c)
    s2 = math.sin(veh2.model.theta_rad[i] - theta_c)

    # translate distances to n-t frame
    dt1 = c1*(impPointEdge[impactNum]['impact_points'][0]) - s1*(impPointEdge[impactNum]['impact_points'][1])
    dn1 = s1*(impPointEdge[impactNum]['impact_points'][0]) + c1*(impPointEdge[impactNum]['impact_points'][1])
    dt2 = c2*(poi_veh2x) - s2*(poi_veh2y)
    dn2 = s2*(poi_veh2x) + c2*(poi_veh2y)

    # translate velocities to n-t frame
    vt1 = c1*veh1.model.vx[i] - s1*veh1.model.vy[i]
    vn1 = s1*veh1.model.vx[i] + c1*veh1.model.vy[i]
    vt2 = c2*veh2.model.vx[i] - s2*veh2.model.vy[i]
    vn2 = s2*veh2.model.vx[i] + c2*veh2.model.vy[i]

    # pre-impact POI velocities (Equations 1)
    vct1 = vt1 - dn1*veh1.model.oz_rad[i]
    vct2 = vt2 - dn2*veh2.model.oz_rad[i]
    vct21 = vct2 - vct1

    vcn1 = vn1 + dt1*veh1.model.oz_rad[i]
    vcn2 = vn2 + dt2*veh2.model.oz_rad[i]
    vcn21 = vcn2 - vcn1

    A11 = (1 / (veh1.weight / 32.2) + 1 / (veh2.weight / 32.2) + dn1**2 / veh1.izz + dn2**2 / veh2.izz)
    A12 = (dt1 * dn1 / veh1.izz + dt2 * dn2 / veh2.izz)
    A22 = (1 / (veh1.weight / 32.2) + 1 / (veh2.weight / 32.2) + dt1**2 / veh1.izz + dt2**2 / veh2.izz)
    A = (A11 * A22 - A12 * A12)

    # collision impulse -
    pt = (1+cor) * (A22 * vct21 + A12 * vcn21) / A
    pn = (1+cor) * (A12 * vct21 + A11 * vcn21) / A

    p_ratio = abs(pt/pn)   # equation 10
    alpha = 1

    if p_ratio > cof:
            print('Sliding Condition')
            pn_s = (1+cor) * vcn21 / (A22 - alpha * cof * A12)
            pt_s = alpha * cof * pn_s

            if pt * pt_s > 0:           # test for sliding direction
                pt = pt_s               # alpha = 1 is correct, keep initial try
                pn = pn_s
            else:
                alpha = -1
                pn = (1+cor) * vcn21 / (A22 - alpha * cof * A12)   # sliding in wrong direction, alpha = -1
                pt = alpha * cof * pn
    else:
        print ('No Sliding')                                                      # no sliding, original pt, pn are correct

    if pn * dn1 < 0:
        print('Compression')
    else:
        print('Tension')

    # change in velocity in collision frame
    # Equation 16 - termination of restitution

    dvt1 = pt / (veh1.weight/32.2)
    dvn1 = pn / (veh1.weight/32.2)
    doz1 = (-1*dn1 * pt + dt1 * pn) / veh1.izz

    dvt2 = -1 * pt / (veh2.weight/32.2)
    dvn2 = -1 * pn / (veh2.weight/32.2)
    doz2 = (dn2 * pt - dt2 * pn) / veh2.izz

    # tranform back to local vehicle coordinate system
    # delta-Vs
    dvx1 = (c1*dvt1 + s1*dvn1)
    dvy1 = (-1*s1*dvt1 + c1*dvn1)
    dveh1 = math.sqrt(dvx1**2 + dvy1**2)

    dvx2 = (c2*dvt2 + s2*dvn2)
    dvy2 = (-1*s2*dvt2 + c2*dvn2)
    dveh2 = math.sqrt(dvx2**2 + dvy2**2)

    # post impact speeds
    vx1_ = veh1.model.vx[i] + dvx1
    vy1_ = veh1.model.vy[i] + dvy1
    veh1_ = math.sqrt(vx1_**2 + vy1_**2)
    oz_rad1_ = doz1 + veh1.model.oz_rad[i]

    vx2_ = veh2.model.vx[i] + dvx2
    vy2_ = veh2.model.vy[i] + dvy2
    veh2_ = math.sqrt(vx2_**2 + vy2_**2)
    oz_rad2_ = doz2 + veh2.model.oz_rad[i]

    # calculations for energy dissipated
    vmt1 = vt1 + dvt1 / (1 + cor)
    omgm1 = veh1.model.oz_rad[i] + doz1 / (1 + cor)
    vmt2 = vt2 + dvt2 / (1 + cor)
    omgm2 = veh2.model.oz_rad[i] + doz2 / (1 + cor)

    vmct21 = vmt2 - vmt1 - dn2 * omgm2 + dn1 * omgm1

    if vmct21 * vct21 < 0:
        print('Reverse Slide')
    else:
        print('Forward Slide')

    vt1_ = vt1 + dvt1
    vt2_ = vt2 + dvt2
    vct21_ = vt2_ - dn2 * oz_rad2_ - vt1_ + dn1 * oz_rad1_

    vn1_ = vn1 + dvn1
    vn2_ = vn2 + dvn2
    vcn21_ = vn2_ + dt2 * oz_rad2_ - vn1_ - dt1 * oz_rad1_

    # dissipated energies
    ke_surface_dis = pt * vmct21
    ke_volume_dis = 0.5 * (1-cor) * (pt * (vct21 - vmct21) + pn * vcn21)
    ke_total_dis = ke_surface_dis + ke_volume_dis

    ke_volume_abs = 0.5 / (1+cor) * (pt * (vct21 - vmct21) + pn * vcn21)
    ke_total_abs = 0.5 / (1+cor) * (pt * (vct21 + vmct21) + pn * vcn21)

    t_effects_dis = 0.5 * pt * (vct21 + vct21_)
    n_effects_dis = 0.5 * pn * (vcn21 + vcn21_)
    tn_total_dis = t_effects_dis + n_effects_dis

    # speeds in fps -
    veh1_impc_result = {'vx_post': vx1_, 'vy_post': vy1_, 'oz_rad_post': oz_rad1_, 'dvx': dvx1, 'dvy': dvy1, 'dv': dveh1}
    veh2_impc_result = {'vx_post': vx2_, 'vy_post': vy2_, 'oz_rad_post': oz_rad2_, 'dvx': dvx2, 'dvy': dvy2, 'dv': dveh2}
    impc_energy = {'t_effects_dis': t_effects_dis, 'n_effects_dis': n_effects_dis, 'tn_total_dis': tn_total_dis}
    impc_results = {'veh1_impc_result': veh1_impc_result, 'veh2_impc_result': veh2_impc_result, 'impc_energy': impc_energy}

    # assign vehicle kinematics to results from IMPC
    # ---- Vehicle 1 ---------- #
    vehicle_list[strikingVehicle].model.vx[i] = vx1_
    vehicle_list[strikingVehicle].model.vy[i] = vy1_
    vehicle_list[strikingVehicle].model.oz_rad[i] = oz_rad1_

    # heading angle
    vehicle_list[strikingVehicle].model.theta_rad[i] = vehicle_list[strikingVehicle].model.theta_rad[i - 1] + dt_motion * oz_rad1_

    # inertial frame coordinates - capital letters
    vehicle_list[strikingVehicle].model.Vx[i] = vx1_ * math.cos(vehicle_list[strikingVehicle].model.theta_rad[i]) - vy1_ * math.sin(vehicle_list[strikingVehicle].model.theta_rad[i])
    vehicle_list[strikingVehicle].model.Vy[i] = vx1_ * math.sin(vehicle_list[strikingVehicle].model.theta_rad[i]) + vy1_ * math.cos(vehicle_list[strikingVehicle].model.theta_rad[i])

    # ------------- Vehicle 2 ------------ #
    vehicle_list[struckVehicle].model.vx[i] = vx2_
    vehicle_list[struckVehicle].model.vy[i] = vy2_
    vehicle_list[struckVehicle].model.oz_rad[i] = oz_rad2_

    # heading angle
    vehicle_list[struckVehicle].model.theta_rad[i] = vehicle_list[struckVehicle].model.theta_rad[i - 1] + dt_motion * oz_rad2_

    # inertial frame coordinates - capital letters
    vehicle_list[struckVehicle].model.Vx[i] = vx2_ * math.cos(vehicle_list[struckVehicle].model.theta_rad[i]) - vy2_ * math.sin(vehicle_list[struckVehicle].model.theta_rad[i])
    vehicle_list[struckVehicle].model.Vy[i] = vx2_ * math.sin(vehicle_list[struckVehicle].model.theta_rad[i]) + vy2_ * math.cos(vehicle_list[struckVehicle].model.theta_rad[i])

    if show_results:
        print("---- IMPC Inputs ----")
        print(f"IMPC for impact: {impactNum} COR: {cor}, COF: {cof}")
        print(f'theta c (deg): {theta_c * 180 / 3.14159:0.1f}')
        print(f"dx1: {impPointEdge[impactNum]['impact_points'][0]}, dy1: {impPointEdge[impactNum]['impact_points'][1]}")
        print(f'dx2: {poi_veh2x}, dy2: {poi_veh2y}')
        print(f'Vx1: {veh1.model.vx[i]:0.2f} fps, Vy1: {veh1.model.vy[i]:0.2f} fps')
        print(f'Vx2: {veh2.model.vx[i]:0.2f} fps, Vy2: {veh2.model.vy[i]:0.2f} fps')
        print(f'Omega1: {veh1.model.oz_rad[i]}')
        print(f'Omega2: {veh2.model.oz_rad[i]}')

        print("")
        print("---- IMPC Results ---- ")
        print(impc_energy)
        print(f'Veh1 DVx: {dvx1 * 0.681818:0.2f} mph, DVy: {dvy1 * 0.681818:0.2f} mph')
        print(f'Veh2 DVx: {dvx2 * 0.681818:0.2f} mph, DVy: {dvy2 * 0.681818:0.2f} mph')

    return vehicle_list, impc_results
""



import numpy as np
import os

"""
calculates intervehicular forces using a sideswipe model (mutual crush + frictional forces)
validated using Funk () see validation directory for simulations and reports
"""


def ss(vehicle_list, crush_data, kmutual, vehicle_mu, i):
    """
    normal force is applied normal to the struck vehicle (vehicle 2)
    frictional force is applied along the impacting plane opposite to relative velocity
    kmutual = [lb/ft]
    """

    print(f'Sideswipe Model Accessed at t = {vehicle_list[0].model.t[i]} seconds')
    # get velocity of impact point on V1 and V2 to relative velocity along impact edge
    # translate to global frame for comparison
    # if relative velocity is zero, friction force is zero

    # instead of relative velocity, try edge location - compare current to prior
    relative_edge_motion = crush_data.edge_loc[i] - crush_data.edge_loc[i - 1]

    # crush force is a function of normal displacement and mutual stiffness
    # normal crush is always negative
    # TODO: need to test each condition

    fmutual = crush_data.normal_crush[i] * kmutual

    # forces in vehicle 2 reference frame
    if vehicle_list[1].edgeimpact == 1:
        vehicle_list[1].model.Fx[i] = fmutual
        vehicle_list[1].model.Fy[i] = - 1 * np.sign(relative_edge_motion) * fmutual * vehicle_mu
        vehicle_list[1].model.Mz[i] = -1 * vehicle_list[1].model.Fx[i] * crush_data.impactp_veh2y[i] + vehicle_list[1].model.Fy[i] * vehicle_list[1].edgeimpact_x1
    elif vehicle_list[1].edgeimpact == 2:
        vehicle_list[1].model.Fx[i] = np.sign(relative_edge_motion) * fmutual * vehicle_mu
        vehicle_list[1].model.Fy[i] = fmutual
        vehicle_list[1].model.Mz[i] = -1 * vehicle_list[1].model.Fx[i] * vehicle_list[1].edgeimpact_y1 + vehicle_list[1].model.Fy[i] * crush_data.impactp_veh2x[i]
    elif vehicle_list[1].edgeimpact == 3:
        vehicle_list[1].model.Fx[i] = -1 * fmutual
        vehicle_list[1].model.Fy[i] = np.sign(relative_edge_motion) * fmutual * vehicle_mu
        vehicle_list[1].model.Mz[i] = -1 * vehicle_list[1].model.Fx[i] * crush_data.impactp_veh2y[i] - vehicle_list[1].model.Fy[i] * vehicle_list[1].edgeimpact_x1
    elif vehicle_list[1].edgeimpact == 4:
        vehicle_list[1].model.Fx[i] = -1 * np.sign(relative_edge_motion) * fmutual * vehicle_mu
        vehicle_list[1].model.Fy[i] = -1 * fmutual
        vehicle_list[1].model.Mz[i] = vehicle_list[1].model.Fx[i] * vehicle_list[1].edgeimpact_y1 + vehicle_list[1].model.Fy[i] * crush_data.impactp_veh2x[i]

    # vehicle 1 forces are equal and opposite to vehicle 2
    # use heading angle to translate forces to vehicle 1 frame
    # can subtract two heading angles - yes
    vehicle_list[0].model.Fx = 0
    vehicle_list[0].model.Fy = 0
    vehicle_list[0].model.Mz = 0

    return vehicle_list



"""
Tire Model - Calculates forces at each tire
accounts for pitch due to braking and roll from cornering forces

Dependencies - v, vx, vy, au, av, omega, constants
"""

import math
import numpy as np
import math

"""
TODO: detailed suspension properties
roll_rate = 6    # semi-firm 7 = semi soft, 3 = extremely firm (corvette) (degrees / g)
rc_cg = 18/12    # passenger car - roll center to cg height (h1)  (ft)
roll_h = 6/12    # roll center height
"""
def sign(x):
    # returns the sign of a number
    if x > 0:
        return 1
    elif x < 0:
        return -1
    elif x == 0:
        return 0


def tire_forces(veh, i, sim_defaults):
    """
    calculate tire forces for the given time step
    """
    # load defaults
    mu_max = sim_defaults['mu_max']  # maximum available friction
    tire_mu = 1
    alpha_max = sim_defaults['alpha_max']  # maximum tire slip angle (rad)
    alpha_max = alpha_max * mu_max
    # TODO: incorporate grade / bank

    if i == 0:
        j = i
    else:
        j = i - 1  # tire forces based on prior time step

    # current steer angle
    veh.model.delta_deg[i] = veh.driver_input.steer[i] / veh.steer_ratio   # steer angle (delta) will always be derived from driver input
    veh.model.delta_rad[i] = veh.model.delta_deg[i] * (math.pi/180)        # net steer angle

    # Ackerman steering
    if veh.model.delta_rad[i] > 0:   # positive steer
        r_net = veh.wb / np.tan(veh.model.delta_rad[i])  # turn radius  (Gillespie)
        rf_steer_angle = np.arctan(veh.wb / (r_net - veh.track))
        lf_steer_angle = veh.model.delta_rad[i]
    elif veh.model.delta_rad[i] < 0:
        r_net = veh.wb / np.tan(veh.model.delta_rad[i])  # turn radius  (Gillespie)
        rf_steer_angle = veh.model.delta_rad[i]
        lf_steer_angle = np.arctan(veh.wb / (r_net + veh.track))
    elif veh.model.delta_rad[i] == 0:
        rf_steer_angle = 0
        lf_steer_angle = 0

    # Forward / Rearward weight shift due to braking or acceleration
    veh_m = veh.weight / 32.2
    veh.model.lf_fz[i] = 0.5 * ((-veh_m * veh.model.au[j] * veh.hcg + veh.weight * veh.lcgr) / veh.wb + veh_m * veh.model.av[j] * veh.hcg / veh.track)
    veh.model.rf_fz[i] = 0.5 * ((-veh_m * veh.model.au[j] * veh.hcg + veh.weight * veh.lcgr) / veh.wb - veh_m * veh.model.av[j] * veh.hcg / veh.track)
    veh.model.rr_fz[i] = 0.5 * ((veh_m * veh.model.au[j] * veh.hcg + veh.weight * veh.lcgf) / veh.wb - veh_m * veh.model.av[j] * veh.hcg / veh.track)
    veh.model.lr_fz[i] = 0.5 * ((veh_m * veh.model.au[j] * veh.hcg + veh.weight * veh.lcgf) / veh.wb + veh_m * veh.model.av[j] * veh.hcg / veh.track)

    # ------------------------ Left Front Tire ----------------------------------- #
    # local velocity
    lf_vx = veh.model.vx[j] + veh.model.oz_rad[j] * (veh.track / 2)
    lf_vy = veh.model.vy[j] + veh.model.oz_rad[j] * veh.lcgf

    veh.model.lf_lock[i] = 0  # locked status of Left Front Wheel - initially set to unlocked
    veh.model.lf_alpha[i] = -1 * np.arctan2(lf_vy, lf_vx) + lf_steer_angle   # tire slip angle (rad)

    if math.fabs(veh.model.lf_alpha[i]) > alpha_max:  # following Steffan 1996 SAE No. 960886
        lf_latf = sign(veh.model.lf_alpha[i]) * mu_max * veh.model.lf_fz[i]  # lateral force if alpha is greater than maximum slip angle - input
    else:
        lf_latf = (veh.model.lf_alpha[i] / alpha_max) * mu_max * veh.model.lf_fz[i]  # lateral force for slip angle less than maximum allowed - input

    if veh.fwd == 1:
        lf_app = veh.model.lf_fz[i] * (mu_max * (veh.driver_input.throttle[i] - veh.driver_input.brake[i] * sign(lf_vx)))  # longitudinal force applied throttle and braking are expressed as % of total friction, will not occur at same time, so add for efficiency
    elif veh.rwd == 1:
        lf_app = -1 * veh.model.lf_fz[i] * (mu_max * veh.driver_input.brake[i] * sign(lf_vx))  # rear wheel drive, front wheel will not apply throttle force
    elif veh.awd == 1:
        lf_app = veh.model.lf_fz[i] * (mu_max * (veh.driver_input.throttle[i] - veh.driver_input.brake[i] * sign(lf_vx)))

    if math.sqrt(lf_app ** 2 + lf_latf ** 2) > (mu_max * veh.model.lf_fz[i]):  # Equation 3 - is the force applied greater than available from friction at tire?
        veh.model.lf_lock[i] = 1
        lf_lonf = -1 * sign(lf_vx) * math.cos(veh.model.lf_alpha[i]) * mu_max * veh.model.lf_fz[i]  # force will be applied in the direction opposite of vehicle motion
        lf_latf = math.sin(veh.model.lf_alpha[i]) * mu_max * veh.model.lf_fz[i]
    elif math.sqrt(lf_app ** 2 + lf_latf ** 2) <= mu_max * veh.model.lf_fz[i]:
        veh.model.lf_lock[i] = 0
        lf_lonf = lf_app
        lf_latf = lf_latf

    # ------------------------ Right Front Tire ----------------------------------- #
    # local velocity
    rf_vx = veh.model.vx[j] - veh.model.oz_rad[j] * (veh.track / 2)
    rf_vy = veh.model.vy[j] + veh.model.oz_rad[j] * veh.lcgf

    veh.model.rf_lock[i] = 0  # locked status - initially set to unlocked
    veh.model.rf_alpha[i] = -1 * np.arctan2(rf_vy, rf_vx) + rf_steer_angle  # tire slip angle (rad)

    if math.fabs(veh.model.rf_alpha[i]) > alpha_max:  # following Steffan 1996 SAE No. 960886
        rf_latf = sign(veh.model.rf_alpha[i]) * mu_max * veh.model.rf_fz[i]  # lateral force if alpha is greater than maximum slip angle - input
    else:
        rf_latf = (veh.model.rf_alpha[i] / alpha_max) * mu_max * veh.model.rf_fz[i]  # lateral force for slip angle less than maximum allowed - input

    # longitudinal Force Applied = f(Vehicle drive tires)
    if veh.fwd == 1:
        rf_app = veh.model.rf_fz[i] * (mu_max * (veh.driver_input.throttle[i] - veh.driver_input.brake[i] * sign(rf_vx)))  # longitudinal force applied throttle and braking are expressed as % of total friction, will not occur at same time, so add for efficiency
    elif veh.rwd == 1:
        rf_app = -1 * veh.model.rf_fz[i] * (mu_max * veh.driver_input.brake[i] * sign(rf_vx))  # rear wheel drive, front wheel will not apply accelerative force
    elif veh.awd == 1:
        rf_app = veh.model.rf_fz[i] * (mu_max * (veh.driver_input.throttle[i]) - veh.driver_input.brake[i] * sign(rf_vx))

    if math.sqrt(rf_app ** 2 + rf_latf ** 2) > mu_max * veh.model.rf_fz[i]:
        veh.model.rf_lock[i] = 1
        rf_lonf = -1 * sign(rf_vx) * math.cos(veh.model.rf_alpha[i]) * mu_max * veh.model.rf_fz[i]
        rf_latf = math.sin(veh.model.rf_alpha[i]) * mu_max * veh.model.rf_fz[i]
    elif math.sqrt(rf_app ** 2 + rf_latf ** 2) <= mu_max * veh.model.rf_fz[i]:
        veh.model.rf_lock[i] = 0
        rf_lonf = rf_app
        rf_latf = rf_latf

    # ------------------------ Right Rear Tire ----------------------------------- #
    # local velocity
    rr_vx = veh.model.vx[j] - veh.model.oz_rad[j] * (veh.track / 2)
    rr_vy = veh.model.vy[j] - veh.model.oz_rad[j] * veh.lcgr

    veh.model.rr_lock[i] = 0  # locked status - initially set to unlocked
    veh.model.rr_alpha[i] = -1 * np.arctan2(rr_vy, rr_vx)  # tire slip angle (rad)

    if math.fabs(veh.model.rr_alpha[i]) > alpha_max:  # following Steffan 1996 SAE No. 960886
        rr_latf = sign(veh.model.rr_alpha[i]) * mu_max * veh.model.rr_fz[i]  # lateral force if alpha is greater than maximum slip angle - input
    else:
        rr_latf = (veh.model.rr_alpha[i] / alpha_max) * mu_max * veh.model.rr_fz[i]  # lateral force for slip angle less than maximum allowed - input

    # longitudinal Force Applied = f(Vehicle drive tires)
    if veh.fwd == 1:
        rr_app = -1 * veh.model.rr_fz[i] * (mu_max * veh.driver_input.brake[i] * sign(rr_vx))  # front wheel drive, rear wheel will not apply accelerative force
    elif veh.rwd == 1:
        rr_app = veh.model.rr_fz[i] * (mu_max * (veh.driver_input.throttle[i] - veh.driver_input.brake[i] * sign(rr_vx)))  # longitudinal force applied throttle and braking are expressed as % of total friction, will not occur at same time, so add for efficiency   (cons['mu_max'] * (veh.driver_input.throttle[i] - veh.driver_input.brake[i]) / 2)
    elif veh.awd == 1:
        rr_app = veh.model.rr_fz[i] * (mu_max * (veh.driver_input.throttle[i] - veh.driver_input.brake[i] * sign(rr_vx)))

    if math.sqrt(rr_app ** 2 + rr_latf ** 2) > mu_max * veh.model.rr_fz[i]:
        veh.model.rr_lock[i] = 1
        rr_lonf = -1 * sign(rr_vx) * math.cos(veh.model.rr_alpha[i]) * mu_max * veh.model.rr_fz[i]
        rr_latf = math.sin(veh.model.rr_alpha[i]) * mu_max * veh.model.rr_fz[i]
    elif math.sqrt(rr_app ** 2 + rr_latf ** 2) <= mu_max * veh.model.rr_fz[i]:
        veh.model.rr_lock[i] = 0
        rr_lonf = rr_app
        rr_latf = rr_latf

    # ------------------------ Left Rear Tire ------------------------------------ #
    lr_vx = veh.model.vx[j] + veh.model.oz_rad[j] * (veh.track / 2)
    lr_vy = veh.model.vy[j] - veh.model.oz_rad[j] * veh.lcgr
    veh.model.lr_lock[i] = 0  # locked status - initially set to unlocked
    veh.model.lr_alpha[i] = -1 * np.arctan2(lr_vy, lr_vx)  # tire slip angle (rad)

    if math.fabs(veh.model.lr_alpha[i]) > alpha_max:  # following Steffan 1996 SAE No. 960886
        lr_latf = sign(veh.model.lr_alpha[i]) * mu_max * veh.model.lr_fz[i]  # lateral force if alpha is greater than maximum slip angle - input
    else:
        lr_latf = (veh.model.lr_alpha[i] / alpha_max) * mu_max * veh.model.lr_fz[i]  # lateral force for slip angle less than maximum allowed - input

    if veh.fwd == 1:
        lr_app = -1 * veh.model.lr_fz[i] * (mu_max * veh.driver_input.brake[i] * sign(lr_vx))  # front wheel drive, rear wheel will not apply accelerative force
    elif veh.rwd == 1:
        lr_app = veh.model.lr_fz[i] * (mu_max * (veh.driver_input.throttle[i] - veh.driver_input.brake[i] * sign(lr_vx)))  # longitudinal force applied throttle and braking are expressed as % of total friction, will not occur at same time, so add for efficiency   (cons['mu_max'] * (veh.driver_input.throttle[i] - veh.driver_input.brake[i]) / 2)
    elif veh.awd == 1:
        lr_app = veh.model.lr_fz[i] * (mu_max * (veh.driver_input.throttle[i] - veh.driver_input.brake[i] * sign(lr_vx)))

    if math.sqrt(lr_app ** 2 + lr_latf ** 2) > mu_max * veh.model.lr_fz[i]:
        veh.model.lr_lock[i] = 1
        lr_lonf = -1 * sign(lr_vx) * math.cos(veh.model.lr_alpha[i]) * mu_max * veh.model.lr_fz[i]
        lr_latf = math.sin(veh.model.lr_alpha[i]) * mu_max * veh.model.lr_fz[i]
    elif math.sqrt(lr_app ** 2 + lr_latf ** 2) <= mu_max * veh.model.lr_fz[i]:
        veh.model.lr_lock[i] = 0
        lr_lonf = lr_app
        lr_latf = lr_latf

    # Calculate vehicle forces in vehicle frame
    # Left Front Tire #
    veh.model.lf_fx[i] = lf_lonf * math.cos(veh.model.delta_rad[i]) - lf_latf * math.sin(veh.model.delta_rad[i])
    veh.model.lf_fy[i] = lf_lonf * math.sin(veh.model.delta_rad[i]) + lf_latf * math.cos(veh.model.delta_rad[i])
    # Right Front Tire #
    veh.model.rf_fx[i] = rf_lonf * math.cos(veh.model.delta_rad[i]) - rf_latf * math.sin(veh.model.delta_rad[i])
    veh.model.rf_fy[i] = rf_lonf * math.sin(veh.model.delta_rad[i]) + rf_latf * math.cos(veh.model.delta_rad[i])
    # Right Rear Tire #
    veh.model.rr_fx[i] = rr_lonf
    veh.model.rr_fy[i] = rr_latf
    # Left Rear Tire #
    veh.model.lr_fx[i] = lr_lonf
    veh.model.lr_fy[i] = lr_latf

    veh.model.lf_lock[i] = veh.model.lf_lock[i].astype(int)
    veh.model.rf_lock[i] = veh.model.rf_lock[i].astype(int)
    veh.model.rr_lock[i] = veh.model.rr_lock[i].astype(int)
    veh.model.lr_lock[i] = veh.model.lr_lock[i].astype(int)

    return veh



from scipy import signal
import matplotlib.pyplot as plt
import numpy as np

def cfcfilt(CFC, data, dt):
    """
    Create filter using CFC based on SAE J211
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.filtfilt.html
    """
    wd = 2 * 3.14159 * CFC * 2.0775
    wa = (np.sin(wd * dt / 2)) / (np.cos(wd * dt / 2))
    aO = (wa**2) / (1 + wa * (2**0.5) + wa**2)
    a1 = 2 * aO
    a2 = aO
    b1 = -2 * ((wa**2) - 1) / (1 + wa * (2**0.5) + wa**2)
    b2 = (-1 + wa * (2**0.5) - wa**2) / (1 + wa * (2**0.5) + wa**2)
    A = [1, -b1, -b2]
    B = [aO, a1, a2]
    filteredData = signal.filtfilt(B, A, data)
    return filteredData




# Calculation Defaults
# ----------------------------------------------------------------------------------------------
"""
configuration for default values and settings
settings effect the following modulues:
vehicle
sdof_calculations
kinematics
kinematicstwo
"""
default_dict = {'dt_motion':0.1,            # time step for vehicle motion (seconds)
                'dt_impact':0.0001,         # time step for impact simulation (seconds)
                'mu_max':0.9,               # maximum available friction
                'alpha_max':0.349066,    # maximum tire slip angle (rad) default is 20 degrees
                'vehicle_mu':0.3}           # intervehicle friction for sideswipe / impc models

# Plot Defaults
# ----------------------------------------------------------------------------------------------
plot_settings = {'figure_size':(16, 9),      # (width, height) for plots
                'plotly_theme':'plotly_white'}  # overall theme for plots



"""
plots used witin vehicle class
"""
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import math

# TODO: create input for figure size - loads from "defaults" folder?
figure_size = (800, 450)

wheel_colors = ['rgb(0, 0, 255)', 'rgb(0, 255, 0)', 'rgb(153, 0, 204)', 'rgb(255, 102, 0)']


def tire_details(veh):
    fig = make_subplots(rows=2, cols=1,
                        shared_xaxes=True,
                        vertical_spacing=0.11)

    # tire center plot depends on lock condition
    def setdash(x):
        if x == 0:
            return 'dash'
        elif x == 1:
            return 'solid'

    # slip angle
    fig.add_trace(go.Scatter(x=veh.model.t, y=veh.model.lf_alpha * 180 / math.pi,
                             mode='lines',
                             name='LF',
                             line=dict(color=wheel_colors[0], width=1,
                                       dash='solid')),
                  row=1, col=1)
    fig.add_trace(go.Scatter(x=veh.model.t, y=veh.model.rf_alpha * 180 / math.pi,
                             mode='lines',
                             name='RF',
                             line=dict(color=wheel_colors[1], width=1,
                                       dash='solid')),
                  row=1, col=1)
    fig.add_trace(go.Scatter(x=veh.model.t, y=veh.model.rr_alpha * 180 / math.pi,
                             mode='lines',
                             name='RR',
                             line=dict(color=wheel_colors[2], width=1,
                                       dash='solid')),
                  row=1, col=1)
    fig.add_trace(go.Scatter(x=veh.model.t, y=veh.model.lr_alpha * 180 / math.pi,
                             mode='lines',
                             name='LR',
                             line=dict(color=wheel_colors[3], width=1,
                                       dash='solid')),
                  row=1, col=1)
    # locked Status
    fig.add_trace(go.Scatter(x=veh.model.t, y=veh.model.lf_lock,
                             showlegend=False,
                             mode='lines',
                             name='LF',
                             line=dict(color=wheel_colors[0], width=1,
                                       dash='solid')),
                  row=2, col=1)

    fig.add_trace(go.Scatter(x=veh.model.t, y=veh.model.rf_lock,
                             showlegend=False,
                             mode='lines',
                             name='RF',
                             line=dict(color=wheel_colors[1], width=1,
                                       dash='solid')),
                  row=2, col=1)

    fig.add_trace(go.Scatter(x=veh.model.t, y=veh.model.rr_lock,
                             showlegend=False,
                             mode='lines',
                             name='RR',
                             line=dict(color=wheel_colors[2], width=1,
                                       dash='solid')),
                  row=2, col=1)

    fig.add_trace(go.Scatter(x=veh.model.t, y=veh.model.lr_lock,
                             showlegend=False,
                             mode='lines',
                             name='LR',
                             line=dict(color=wheel_colors[3], width=1,
                                       dash='solid')),
                  row=2, col=1)

    fig.update_layout(
        legend=dict(orientation="h", yanchor='top', y=1.15, xanchor='left', x=0.01),
        autosize=False,
        width=figure_size[0],
        height=figure_size[1],
        title='Tire Slip Angle & Saturation (locked) Status',
        template='plotly_white',
        font=dict(family='Arial', size=16, color='black'))

    fig.update_xaxes(showline=True, linewidth=1, linecolor='black', ticks="outside",
                     tickwidth=1, tickcolor='black', ticklen=10, zeroline=False)
    fig.update_yaxes(showline=True, linewidth=1, linecolor='black', ticks="outside",
                     tickwidth=1, tickcolor='black', ticklen=10, zeroline=False)
    fig.update_xaxes(showgrid=False, title_text='Time (s)', row=2, col=1)
    fig.update_xaxes(showgrid=False, title_text='', row=1, col=1)
    fig.update_yaxes(showgrid=False, title_text='Locked Satus', row=2, col=1)
    fig.update_yaxes(showgrid=False, title_text='Slip Angle (degrees)', row=1, col=1)

    fig.show()


def vertical_forces(veh):
    fig = go.Figure()

    fig.add_trace(go.Scatter(x=veh.model.t, y=veh.model.lf_fz,
                             mode='lines',
                             name='LF',
                             line=dict(color=wheel_colors[0], width=1,
                                       dash='solid')))
    fig.add_trace(go.Scatter(x=veh.model.t, y=veh.model.rf_fz,
                             mode='lines',
                             name='RF',
                             line=dict(color=wheel_colors[1], width=1,
                                       dash='solid')))
    fig.add_trace(go.Scatter(x=veh.model.t, y=veh.model.rr_fz,
                             mode='lines',
                             name='RR',
                             line=dict(color=wheel_colors[2], width=1,
                                       dash='solid')))
    fig.add_trace(go.Scatter(x=veh.model.t, y=veh.model.lr_fz,
                             mode='lines',
                             name='LR',
                             line=dict(color=wheel_colors[3], width=1,
                                       dash='solid')))
    fig.update_layout(
        legend=dict(orientation="h", yanchor='top', y=1.15, xanchor='left', x=0.01),
        autosize=False,
        width=figure_size[0],
        height=figure_size[1],
        title='Vertical Tire Forces',
        template='plotly_white',
        font=dict(family='Arial', size=16, color='black'))

    fig.update_xaxes(showline=True, linewidth=1, linecolor='black', ticks="outside",
                     tickwidth=1, tickcolor='black', ticklen=10, zeroline=False, title_text='Time (s)')
    fig.update_yaxes(showline=True, linewidth=1, linecolor='black', ticks="outside",
                     tickwidth=1, tickcolor='black', ticklen=10, zeroline=False, title_text='Vertical Force (lb)')

    fig.show()


def long_forces(veh):
    fig = go.Figure()

    fig.add_trace(go.Scatter(x=veh.model.t, y=veh.model.lf_fx,
                             mode='lines',
                             name='LF',
                             line=dict(color=wheel_colors[0], width=1,
                                       dash='solid')))
    fig.add_trace(go.Scatter(x=veh.model.t, y=veh.model.rf_fx,
                             mode='lines',
                             name='RF',
                             line=dict(color=wheel_colors[1], width=1,
                                       dash='solid')))
    fig.add_trace(go.Scatter(x=veh.model.t, y=veh.model.rr_fx,
                             mode='lines',
                             name='RR',
                             line=dict(color=wheel_colors[2], width=1,
                                       dash='solid')))
    fig.add_trace(go.Scatter(x=veh.model.t, y=veh.model.lr_fx,
                             mode='lines',
                             name='LR',
                             line=dict(color=wheel_colors[3], width=1,
                                       dash='solid')))
    fig.update_layout(
        legend=dict(orientation="h", yanchor='top', y=1.15, xanchor='left', x=0.01),
        autosize=False,
        width=figure_size[0],
        height=figure_size[1],
        title='Forward Tire Forces',
        template='plotly_white',
        font=dict(family='Arial', size=16, color='black'))

    fig.update_xaxes(showline=True, linewidth=1, linecolor='black', ticks="outside",
                     tickwidth=1, tickcolor='black', ticklen=10, zeroline=False, title_text='Time (s)')
    fig.update_yaxes(showline=True, linewidth=1, linecolor='black', ticks="outside",
                     tickwidth=1, tickcolor='black', ticklen=10, zeroline=False, title_text='Forward Force (lb)')

    fig.show()
